from torch import nn
import torch.nn.functional as F
from flgo.utils.fmodule import FModule

class Model(FModule):
    def __init__(self):
        super().__init__()
        self.conv1 = nn.Conv2d(in_channels=1, out_channels=32, kernel_size=5, padding=2)
        self.conv2 = nn.Conv2d(in_channels=32, out_channels=64, kernel_size=5, padding=2)
        self.fc1 = nn.Linear(3136, 512)
        self.fc2 = nn.Linear(512, 62)

    def forward(self, x):
        x = self.get_embedding(x)
        x = self.fc2(x)
        return x

    def get_embedding(self, x):
        x = x.view((x.shape[0], 28, 28))
        x = x.unsqueeze(1)
        x = F.max_pool2d(F.relu(self.conv1(x)), 2)
        x = F.max_pool2d(F.relu(self.conv2(x)), 2)
        x = x.view(-1, x.shape[1] * x.shape[2] * x.shape[3])
        x = F.relu(self.fc1(x))
        return x

def init_local_module(object):
    pass

def init_global_module(object):
    if 'Server' in object.__class__.__name__:
        object.model = Model().to(object.device)


"""This is a non-official implementation of 'Asynchronous Federated Optimization' (http://arxiv.org/abs/1903.03934). """
from .fedbase import BasicServer
from .fedbase import BasicClient as Client

class Server(BasicServer):
    def initialize(self):
        self.init_algo_para({'period':20, 'alpha': 0.6, 'mu':0.005, 'flag':'constant', 'hinge_a':10, 'hinge_b':6, 'poly_a':0.5})
        self.tolerance_for_latency = 1000
        self.client_taus = [0 for _ in self.clients]

    def iterate(self):
        # Scheduler periodically triggers the idle clients to locally train the model
        self.selected_clients = self.sample() if (self.gv.clock.current_time%self.period)==0 or self.gv.clock.current_time==1 else []
        if len(self.selected_clients)>0:
            self.gv.logger.info('Select clients {} at time {}'.format(self.selected_clients, self.gv.clock.current_time))
        # Record the timestamp of the selected clients
        for cid in self.selected_clients: self.client_taus[cid] = self.current_round
        # Check the currently received models
        res = self.communicate(self.selected_clients, asynchronous=True)
        received_models = res['model']
        received_client_ids = res['__cid']
        if len(received_models) > 0:
            self.gv.logger.info('Receive new models from clients {} at time {}'.format(received_client_ids, self.gv.clock.current_time))
            # averaging the simultaneously received models at the current moment
            taus = [self.client_taus[cid] for cid in received_client_ids]
            alpha_ts = [self.alpha * self.s(self.current_round - tau) for tau in taus]
            currently_updated_models = [(1-alpha_t)*self.model+alpha_t*model_k for alpha_t, model_k in zip(alpha_ts, received_models) ]
            self.model = self.aggregate(currently_updated_models)
        return len(received_models) > 0

    def s(self, delta_tau):
        if self.flag == 'constant':
            return 1
        elif self.flag == 'hinge':
            return 1 if delta_tau <= self.b else 1.0 / (self.a * (delta_tau - self.b))
        elif self.flag == 'poly':
            return (delta_tau + 1) ** (-self.a)



import copy

import community.community_louvain
import networkx as nx
import torch_geometric
from torch_geometric.transforms import RandomLinkSplit
from torch_geometric.utils import negative_sampling, from_networkx
import torch_geometric.utils
import collections

from flgo.benchmark.toolkits.base import *


class LinkPredicitonTaskGenenerator(BasicTaskGenerator):
    def __init__(self, benchmark, rawdata_path, builtin_class, dataset_name, transforms=None, num_clients=10):
        super(LinkPredicitonTaskGenenerator, self).__init__(benchmark, rawdata_path)
        self.builtin_class = builtin_class
        self.dataset_name = dataset_name
        self.transforms = transforms
        self.num_clients = num_clients

    def load_data(self):
        all_data = self.builtin_class(root=self.rawdata_path, name=self.dataset_name, transform=self.transforms).data
        transform = RandomLinkSplit(is_undirected=all_data.is_undirected(), num_test=0.2, num_val=0,
                                    add_negative_train_samples=False)
        local_data, _, test_data = transform(all_data)
        self.local_data_edge_label_index = local_data.edge_label_index
        self.test_data_edge_label_index = test_data.edge_label_index
        self.test_data_edge_label = test_data.edge_label
        self.train_G = torch_geometric.utils.to_networkx(all_data, to_undirected=all_data.is_undirected(),
                                                         node_attrs=['x'])

    def partition(self):
        self.local_nodes = [[] for _ in range(self.num_clients)]
        node_groups = community.community_louvain.best_partition(self.train_G)
        groups = collections.defaultdict(list)
        for ni, gi in node_groups.items():
            groups[gi].append(ni)
        groups = {k: groups[k] for k in list(range(len(groups)))}
        # ensure the number of groups is larger than the number of clients
        while len(groups) < self.num_clients:
            # find the group with the largest size
            groups_lens = [groups[k] for k in range(len(groups))]
            max_gi = np.argmax(groups_lens)
            # set the size of the new group
            min_glen = min(groups_lens)
            max_glen = max(groups_lens)
            if max_glen < 2 * min_glen: min_glen = max_glen // 2
            # split the group with the largest size into two groups
            nodes_in_gi = groups[max_gi]
            new_group_id = len(groups)
            groups[new_group_id] = nodes_in_gi[:min_glen]
            groups[max_gi] = nodes_in_gi[min_glen:]
        # allocate different groups to clients
        groups_lens = [groups[k] for k in range(len(groups))]
        group_ids = np.argsort(groups_lens)
        for gi in group_ids:
            cid = np.argmin([len(li) for li in self.local_nodes])
            self.local_nodes[cid].extend(groups[gi])

    def get_task_name(self):
        return '_'.join(['B-' + self.benchmark, 'P-None', 'N-' + str(self.num_clients)])


class LinkPredicitonTaskPipe(BasicTaskPipe):
    def __init__(self, task_name, buildin_class, transform=None):
        super(LinkPredicitonTaskPipe, self).__init__(task_name)
        self.builtin_class = buildin_class
        self.transform = transform

    def save_task(self, generator):
        client_names = self.gen_client_names(len(generator.local_nodes))
        feddata = {'client_names': client_names,
                   'server_data':
                       {'edge_label_index': generator.test_data_edge_label_index.tolist(),
                        'edge_label': generator.test_data_edge_label.tolist()},
                   'rawdata_path': generator.rawdata_path,
                   'dataset_name': generator.dataset_name,
                   'local_edge_label_index': generator.local_data_edge_label_index.tolist()}
        for cid in range(len(client_names)): feddata[client_names[cid]] = {'data': generator.local_nodes[cid]}
        with open(os.path.join(self.task_path, 'data.json'), 'w') as outf:
            json.dump(feddata, outf)
        return

    def load_data(self, running_time_option) -> dict:
        # load the datasets
        all_data = self.builtin_class(root=self.feddata['rawdata_path'], name=self.feddata['dataset_name'],
                                           transform=self.transform).data
        server_data = copy.deepcopy(all_data)
        server_data.edge_label_index = torch.Tensor(self.feddata['server_data']['edge_label_index'])
        server_data.edge_label = torch.Tensor(self.feddata['server_data']['edge_label'])

        local_data = copy.deepcopy(all_data)
        local_data_edge_label_index = torch.Tensor(self.feddata['local_edge_label_index'])
        edge_index = torch.cat([local_data_edge_label_index,
                                           torch.flip(local_data_edge_label_index, dims=[0])], dim=1)
        local_data.edge_index = edge_index
        server_data.edge_index = edge_index
        task_data = {'server': {'test': server_data}}
        G = torch_geometric.utils.to_networkx(local_data, to_undirected=local_data.is_undirected(),
                                              node_attrs=['x'])
        # rearrange data for clients
        for cid, cname in enumerate(self.feddata['client_names']):
            cdata = from_networkx(nx.subgraph(G, self.feddata[cname]['data']))
            transform = RandomLinkSplit(is_undirected=cdata.is_undirected(), num_test=running_time_option['train_holdout'], num_val=0,
                                        add_negative_train_samples=False)
            train_data, _, val_data = transform(cdata)
            task_data[cname] = {'train': train_data, 'valid': val_data}
        return task_data


class LinkPredicitonTaskCalculator(BasicTaskCalculator):
    def __init__(self, device, optimizer_name='sgd'):
        super(LinkPredicitonTaskCalculator, self).__init__(device, optimizer_name)
        self.criterion = torch.nn.BCEWithLogitsLoss()
        self.DataLoader = torch_geometric.loader.DataLoader


    def compute_loss(self, model, data):
        tdata = self.data_to_device(data)
        z = model.encode(tdata.x, tdata.edge_label_index)
        neg_edge_index = negative_sampling(
            edge_index=tdata.edge_index, num_nodes=tdata.num_nodes,
            num_neg_samples=tdata.edge_label_index.size(1)
        )
        edge_label_index = torch.cat(
            [tdata.edge_label_index, neg_edge_index],
            dim=-1,
        )
        edge_label = torch.cat([
            tdata.edge_label,
            tdata.edge_label.new_zeros(neg_edge_index.size(1))
        ], dim=0)
        outputs = model.decode(z, edge_label_index).view(-1)
        loss = self.criterion(outputs, edge_label)
        return {'loss': loss}

    @torch.no_grad()
    def test(self, model, dataset, batch_size=64, num_workers=0):
        total_loss = 0
        total_num_samples = 0
        tdata = self.data_to_device(dataset)
        z = model.encode(tdata.x, tdata.edge_index)
        outputs = model.decode(z, tdata.edge_label_index).view(-1)
        loss = self.criterion(outputs, tdata.edge_label)
        num_samples = len(tdata.x)
        total_loss += num_samples * loss
        total_num_samples += num_samples
        total_loss = total_loss.item()
        return {'loss': total_loss / total_num_samples}

    def data_to_device(self, data):
        return data.to(self.device)

    def get_dataloader(self, dataset, batch_size=64, shuffle=True, num_workers=0):
        return self.DataLoader([dataset], batch_size=batch_size, shuffle=shuffle)



from torch import nn
from flgo.utils.fmodule import FModule
import torch.nn.functional as F

class Model(FModule):
    def __init__(self):
        super(Model, self).__init__()
        self.conv1 = nn.Conv2d(in_channels=3, out_channels=64, kernel_size=5)
        self.conv2 = nn.Conv2d(in_channels=64, out_channels=64, kernel_size=5)
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)
        self.fc1 = nn.Linear(1600, 384)
        self.fc2 = nn.Linear(384, 192)
        self.fc3 = nn.Linear(192, 100)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = x.view(-1, 1600)
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x

def init_local_module(object):
    pass

def init_global_module(object):
    if 'Server' in object.__class__.__name__:
        object.model = Model().to(object.device)


from torch import nn
from flgo.utils.fmodule import FModule

class Model(FModule):
    def __init__(self, dim_in=3*32*32, dim_hidden=200, dim_out=10):
        super(Model, self).__init__()
        self.layer_input = nn.Linear(dim_in, dim_hidden)
        self.relu = nn.ReLU()
        self.fc1 = nn.Linear(dim_hidden, dim_hidden)
        self.fc2 = nn.Linear(dim_hidden, dim_out)

    def forward(self, x):
        x = self.get_embedding(x)
        x = self.fc2(x)
        return x

    def get_embedding(self, x):
        x = x.view(-1, x.shape[1] * x.shape[-2] * x.shape[-1])
        x = self.layer_input(x)
        x = self.relu(x)
        x = self.fc1(x)
        x = self.relu(x)
        return x

def init_local_module(object):
    pass

def init_global_module(object):
    if 'Server' in object.__class__.__name__:
        object.model = Model().to(object.device)


r"""
This module is to simulate arbitrary system heterogeneity that may occur in practice.
We conclude four types of system heterogeneity from existing works:
    1) availability: the devices will be either available or unavailable at each moment, where only the
                    available devices can be selected to participate in training.
    2) responsiveness: the responsiveness describes the length of the period from the server broadcasting the
                    gloabl model to the server receiving the locally trained model from a particular client.
    3) completeness: since the server cannot fully control the behavior of devices,it's possible for devices to
                    upload imcomplete model updates (i.e. only training for a few steps).
    4) connectivity: the clients who promise to complete training may suffer accidients so that the server may lose
                    connections with these client who will never return the currently trained local model.
We build up a client state machine to simulate the four types of system heterogeneity, and provide high-level
APIs to allow customized system heterogeneity simulation.

Example 1: How to customize the system heterogeneity?
>>> class MySimulator(flgo.system_simulator.base.BasicSimulator):
...     def update_client_availability(self):
...         # update the variable 'prob_available' and 'prob_unavailable' for all the clients
...         self.set_variable(self.all_clients, 'prob_available', [0.9 for _ in self.all_clients])
...         self.set_variable(self.all_clients, 'prob_unavailable', [0.1 for _ in self.all_clients])
...
...     def update_client_connectivity(self, client_ids):
...         # update the variable 'prob_drop' for clients in client_ids
...         self.set_variable(client_ids, 'prob_drop', [0.1 for _ in client_ids])
...
...     def update_client_responsiveness(self, client_ids, *args, **kwargs):
...         # update the variable 'latency' for clients in client_ids
...         self.set_variable(client_ids, 'latency', [np.random.randint(5,100) for _ in client_ids])
...
...     def update_client_completeness(self, client_ids, *args, **kwargs):
...         # update the variable 'working_amount' for clients in client_ids
...         self.set_variable(client_ids, 'working_amount',  [max(int(self.clients[cid].num_steps*np.random.rand()), 1) for cid in client_ids])
>>> r = flgo.init(task, algorithm=fedavg, Simulator=MySimulator)
>>> # The runner r will be runned under the customized system heterogeneity, where the clients' states will be flushed by
>>> # MySimulator.update_client_xxx at each moment of the virtual clock or particular events happen (i.e. a client was selected)

We also provide some preset Simulator like flgo.system_simulator.DefaultSimulator and flgo.system_simulator.
"""
from flgo.system_simulator.default_simulator import Simulator as DefaultSimulator
from flgo.system_simulator.phone_simulator import Simulator as PhoneSimulator



import community.community_louvain
import torch_geometric.transforms as T
from torch_geometric.utils import mask_to_index, index_to_mask, from_networkx
import torch_geometric.utils
import collections
from flgo.benchmark.toolkits.base import *
import networkx as nx


class NodeClassificationTaskGen(BasicTaskGenerator):
    def __init__(self, benchmark, rawdata_path, builtin_class, dataset_name, transforms=None, num_clients=10):
        super(NodeClassificationTaskGen, self).__init__(benchmark, rawdata_path)
        self.builtin_class = builtin_class
        self.dataset_name = dataset_name
        self.transforms = transforms
        self.num_clients = num_clients


    def load_data(self):
        self.all_data = T.RandomNodeSplit(split='train_rest', num_val=0.1, num_test=0.2)(
            self.builtin_class(root=self.rawdata_path, name=self.dataset_name, transform=self.transforms).data)
        self.test_nodes = mask_to_index(self.all_data.test_mask)
        self.all_train_nodes = mask_to_index(self.all_data.train_mask)
        self.all_val_nodes = mask_to_index(self.all_data.val_mask)
        self.G = torch_geometric.utils.to_networkx(self.all_data, to_undirected=self.all_data.is_undirected(),
                                                   node_attrs=['x', 'y', 'train_mask', 'val_mask', 'test_mask'])


    def partition(self):
        self.local_nodes = [[] for _ in range(self.num_clients)]
        node_groups = community.community_louvain.best_partition(self.G)
        groups = collections.defaultdict(list)
        for ni, gi in node_groups.items():
            groups[gi].append(ni)
        groups = {k: groups[k] for k in list(range(len(groups)))}
        # ensure the number of groups is larger than the number of clients
        while len(groups) < self.num_clients:
            # find the group with the largest size
            groups_lens = [groups[k] for k in range(len(groups))]
            max_gi = np.argmax(groups_lens)
            # set the size of the new group
            min_glen = min(groups_lens)
            max_glen = max(groups_lens)
            if max_glen < 2 * min_glen: min_glen = max_glen // 2
            # split the group with the largest size into two groups
            nodes_in_gi = groups[max_gi]
            new_group_id = len(groups)
            groups[new_group_id] = nodes_in_gi[:min_glen]
            groups[max_gi] = nodes_in_gi[min_glen:]
        # allocate different groups to clients
        groups_lens = [groups[k] for k in range(len(groups))]
        group_ids = np.argsort(groups_lens)
        for gi in group_ids:
            cid = np.argmin([len(li) for li in self.local_nodes])
            self.local_nodes[cid].extend(groups[gi])

    def get_task_name(self):
        return '_'.join(['B-'+self.benchmark,  'P-None', 'N-'+str(self.num_clients)])


class NodeClassificationTaskPipe(BasicTaskPipe):
    def __init__(self, task_name, buildin_class, transform=None):
        super(NodeClassificationTaskPipe, self).__init__(task_name)
        self.builtin_class = buildin_class
        self.transform = transform

    def save_task(self, generator):
        client_names = self.gen_client_names(len(generator.local_nodes))
        feddata = {'client_names': client_names,
                   'server_data': generator.test_nodes.tolist(),
                   'rawdata_path': generator.rawdata_path,
                   'dataset_name': generator.dataset_name,
                   'all_train_nodes': generator.all_train_nodes.tolist(),
                   'all_val_nodes': generator.all_val_nodes.tolist()}
        for cid in range(len(client_names)): feddata[client_names[cid]] = {'data': generator.local_nodes[cid]}
        with open(os.path.join(self.task_path, 'data.json'), 'w') as outf:
            json.dump(feddata, outf)
        return

    def load_data(self, running_time_option) -> dict:
        # load the datasets
        self.all_data = self.builtin_class(root=self.feddata['rawdata_path'], name=self.feddata['dataset_name'],
                                           transform=self.transform).data
        self.all_data.test_mask = index_to_mask(torch.LongTensor(self.feddata['server_data']),
                                                size=self.all_data.num_nodes)
        self.all_data.train_mask = index_to_mask(torch.LongTensor(self.feddata['all_train_nodes']),
                                                 size=self.all_data.num_nodes)
        self.all_data.val_mask = index_to_mask(torch.LongTensor(self.feddata['all_val_nodes']),
                                               size=self.all_data.num_nodes)
        task_data = {'server': {'test': self.all_data}}
        G = torch_geometric.utils.to_networkx(self.all_data, to_undirected=self.all_data.is_undirected(),
                                              node_attrs=['x', 'y', 'train_mask', 'val_mask', 'test_mask'])
        # rearrange data for clients
        for cid, cname in enumerate(self.feddata['client_names']):
            cdata = from_networkx(nx.subgraph(G, self.feddata[cname]['data']))
            cdata.test_mask = cdata.val_mask
            task_data[cname] = {'train': cdata, 'valid': cdata}
        return task_data

"""
load_data -> return task_data = {
    'server': {'test': anything, 'xxx': anything},
    'Client01': {'train': anything, ...}
    ...
}

generate_objects -> [object1, object2, ...]
object1.name = task_data[0] = 'server'
object2.name = task_data[1] = 'Client01'
....

distribute:
    specify the object according to the name of the object
        object_x
        x_data = task_data[x_name] (i.e. {'xxx':anything, ...})
        for key in x_data:
            object_x.set_data(key, x_data[key])
            
set_data(data_name, data):
    setattr(self, data_name+'_data', data)
    
"""
class NodeClassificationTaskCalculator(BasicTaskCalculator):
    def __init__(self, device, optimizer_name='sgd'):
        super(NodeClassificationTaskCalculator, self).__init__(device, optimizer_name)
        self.criterion = torch.nn.NLLLoss()
        self.DataLoader = torch_geometric.loader.DataLoader

    def compute_loss(self, model, data):
        tdata = self.data_to_device(data)
        outputs = model(tdata)
        loss = self.criterion(outputs[tdata.train_mask], tdata.y[tdata.train_mask])
        return {'loss': loss}

    @torch.no_grad()
    def test(self, model, dataset, batch_size=64, num_workers=0):
        loader = self.DataLoader([dataset], batch_size=batch_size)
        total_loss = 0
        total_num_samples = 0
        for batch in loader:
            tdata = self.data_to_device(batch)
            outputs = model(tdata)
            loss = self.criterion(outputs[tdata.test_mask], tdata.y[tdata.test_mask])
            num_samples = len(tdata.x)
            total_loss += num_samples * loss
            total_num_samples += num_samples
        total_loss = total_loss.item()
        return {'loss': total_loss / total_num_samples}

    def data_to_device(self, data):
        return data.to(self.device)

    def get_dataloader(self, dataset, batch_size=64, shuffle=True, num_workers=0):
        return self.DataLoader([dataset], batch_size=batch_size, shuffle=shuffle)



from torch_geometric.datasets import Planetoid
import flgo.benchmark
import os.path
from flgo.benchmark.toolkits.graph.horizontal.node_classification import *

TaskCalculator = NodeClassificationTaskCalculator
class TaskGenerator(NodeClassificationTaskGen):
    def __init__(self, rawdata_path=os.path.join(flgo.benchmark.path,'RAW_DATA', 'CITESEER'), *args, **kwargs):
        super(TaskGenerator, self).__init__(benchmark='citeseer_node_classification',
                                            rawdata_path=rawdata_path, builtin_class=Planetoid,
                                            dataset_name='CITESEER', num_clients=10)

class TaskPipe(NodeClassificationTaskPipe):
    def __init__(self, task_name):
        super(TaskPipe, self).__init__(task_name, Planetoid, transform=None)


import collections
import copy
import multiprocessing
import signal
import time
try:
    from collections import Iterable
except ImportError:
    from collections.abc import Iterable
import numpy as np
import torch
import os.path
import json
import flgo.system_simulator
import flgo.system_simulator.default_simulator
import flgo.system_simulator.base
import flgo.utils.fmodule
import flgo.experiment.logger.simple_logger
import flgo.experiment.logger.tune_logger
import flgo.experiment.logger
import flgo.experiment.device_scheduler
from flgo.system_simulator.base import BasicSimulator
import flgo.benchmark.toolkits.partition
import flgo.algorithm
import itertools
import argparse
import importlib
import random
import os
import yaml
import queue
import sys

sample_list=['uniform', 'md', 'full', 'uniform_available', 'md_available', 'full_available']
agg_list=['uniform', 'weighted_scale', 'weighted_com']
optimizer_list=['SGD', 'Adam', 'RMSprop', 'Adagrad']
default_option_dict = {'pretrain': '', 'sample': 'md', 'aggregate': 'uniform', 'num_rounds': 20, 'proportion': 0.2, 'learning_rate_decay': 0.998, 'lr_scheduler': -1, 'early_stop': -1, 'num_epochs': 5, 'num_steps': -1, 'learning_rate': 0.1, 'batch_size': 64.0, 'optimizer': 'SGD', 'momentum': 0, 'weight_decay': 0, 'algo_para': [], 'train_holdout': 0.1, 'test_holdout': 0.0, 'seed': 0, 'gpu': [], 'server_with_cpu': False, 'num_parallels': 1, 'num_workers': 0, 'test_batch_size': 512, 'simulator': 'default_simulator', 'availability': 'IDL', 'connectivity': 'IDL', 'completeness': 'IDL', 'responsiveness': 'IDL', 'logger': 'basic_logger', 'log_level': 'INFO', 'log_file': False, 'no_log_console': False, 'no_overwrite': False, 'eval_interval': 1}

class GlobalVariable:
    """this class is to create a buffer space for sharing variables across different parties for each runner respectively in a single machine"""
    def __init__(self):
        self.logger = None
        self.simulator = None
        self.clock = None
        self.dev_list = None
        self.TaskCalculator = None
        self.TaskPipe = None
        self.crt_dev = 0

    def apply_for_device(self):
        """apply for a new device from currently available ones (i.e. devices in self.dev_list)"""
        if self.dev_list is None: return None
        dev = self.dev_list[self.crt_dev]
        self.crt_dev = (self.crt_dev + 1) % len(self.dev_list)
        return dev

def setup_seed(seed):
    """fix all the random seed used in numpy, torch and random module"""
    random.seed(1+seed)
    np.random.seed(21+seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    torch.manual_seed(12+seed)
    torch.cuda.manual_seed_all(123+seed)
    torch.backends.cudnn.enabled = False
    torch.backends.cudnn.deterministic = True

def read_option_from_command():
    """load configuration for flgo.init from command lines"""
    parser = argparse.ArgumentParser()
    """Training Options"""
    # basic settings
    # methods of server side for sampling and aggregating
    parser.add_argument('--sample', help='methods for sampling clients', type=str, choices=sample_list, default='md')
    parser.add_argument('--aggregate', help='methods for aggregating models', type=str, choices=agg_list, default='uniform')
    # hyper-parameters of training in server side
    parser.add_argument('--num_rounds', help='number of communication rounds', type=int, default=20)
    parser.add_argument('--proportion', help='proportion of clients sampled per round', type=float, default=0.2)
    parser.add_argument('--learning_rate_decay', help='learning rate decay for the training process;', type=float, default=0.998)
    parser.add_argument('--lr_scheduler', help='type of the global learning rate scheduler', type=int, default=-1)
    parser.add_argument('--early_stop', help='stop training if there is no improvement for no smaller than the maximum rounds', type=int, default=-1)
    # hyper-parameters of local training
    parser.add_argument('--num_epochs', help='number of epochs when clients trainset on data;', type=int, default=5)
    parser.add_argument('--num_steps', help='the number of local steps, which dominate num_epochs when setting num_steps>0', type=int, default=-1)
    parser.add_argument('--learning_rate', help='learning rate for inner solver;', type=float, default=0.1)
    parser.add_argument('--batch_size', help='batch size when clients trainset on data;', type=float, default='64')
    parser.add_argument('--optimizer', help='select the optimizer for gd', type=str, choices=optimizer_list, default='SGD')
    parser.add_argument('--momentum', help='momentum of local update', type=float, default=0)
    parser.add_argument('--weight_decay', help='weight decay for the training process', type=float, default=0)
    # algorithm-dependent hyper-parameters
    parser.add_argument('--algo_para', help='algorithm-dependent hyper-parameters', nargs='*', type=float)

    """Environment Options"""
    # the ratio of the amount of the data used to train
    parser.add_argument('--train_holdout', help='the rate of holding out the validation dataset from all the local training datasets', type=float, default=0.1)
    parser.add_argument('--test_holdout', help='the rate of holding out the validation dataset from the training datasets', type=float, default=0.0)
    # realistic machine config
    parser.add_argument('--seed', help='seed for random initialization;', type=int, default=0)
    parser.add_argument('--gpu', nargs='*', help='GPU IDs and empty input is equal to using CPU', type=int)
    parser.add_argument('--server_with_cpu', help='seed for random initialization;', action="store_true", default=False)
    parser.add_argument('--num_parallels', help="the number of parallels in the clients computing session", type=int, default=1)
    parser.add_argument('--num_workers', help='the number of workers of DataLoader', type=int, default=0)
    parser.add_argument('--test_batch_size', help='the batch_size used in testing phase;', type=int, default=512)

    """Simulator Options"""
    # the simulating systemic configuration of clients and the server that helps constructing the heterogeity in the network condition & computing power
    parser.add_argument('--availability', help="client availability mode", type=str, default = 'IDL')
    parser.add_argument('--connectivity', help="client connectivity mode", type=str, default = 'IDL')
    parser.add_argument('--completeness', help="client completeness mode", type=str, default = 'IDL')
    parser.add_argument('--responsiveness', help="client responsiveness mode", type=str, default='IDL')

    """Logger Options"""
    # logger setting
    parser.add_argument('--log_level', help='the level of logger', type=str, default='INFO')
    parser.add_argument('--log_file', help='bool controls whether log to file and default value is False', action="store_true", default=False)
    parser.add_argument('--no_log_console', help='bool controls whether log to screen and default value is True', action="store_true", default=False)
    parser.add_argument('--no_overwrite', help='bool controls whether to overwrite the old result', action="store_true", default=False)
    parser.add_argument('--eval_interval', help='evaluate every __ rounds;', type=int, default=1)
    try: option = vars(parser.parse_known_args()[0])
    except IOError as msg: parser.error(str(msg))
    for key in option.keys():
        if option[key] is None:
            option[key]=[]
    return option

def load_configuration(config={}):
    """load configuration for yml file or dict"""
    if type(config) is str and config.endswith('.yml'):
        with open(config) as f:
            option = yaml.load(f, Loader=yaml.FullLoader)
        return option
    elif type(config) is dict:
        return config

def gen_task_by_para(benchmark, bmk_para:dict={}, Partitioner=None, par_para:dict={}, task_path: str='', rawdata_path:str='', seed:int=0):
    r"""
    Generate a federated task according to the parameters of this function. The formats and meanings of the inputs are listed as below:
    :param
        benchmark (python package || str): the benchmark package or the module path of it
        bmk_para (dict): the customized parameter dict of the method TaskGenerator.__init__() of the benchmark
        Partitioner (class || str): the class of the Partitioner or the name of the Partitioner that was realized in flgo.benchmark.toolkits.partition
        par_para (dict): the customized parameter dict of the method Partitioner.__init__()
        task_path (str): the path to store the generated task
        rawdata_path (str): where the raw data will be downloaded\stored
        seed (int): the random seed used to generate the task
    :return

    Example:
        >>> import flgo
        >>> import flgo.benchmark.mnist_classification as mnist
        >>> from flgo.benchmark.toolkits.partition import IIDPartitioner
        >>> # GENERATE TASK BY PASSING THE MODULE OF BENCHMARK AND THE CLASS OF THE PARTITIOENR
        >>> flgo.gen_task_by_para(benchmark=mnist, Partitioner = IIDPartitioner, par_para={'num_clients':100}, task_path='./mnist_gen_by_para1')
        >>> # GENERATE THE SAME TASK BY PASSING THE STRING
        >>> flgo.gen_task_by_para(benchmark='flgo.benchmark.mnist_classification', Partitioner='IIDPartitioner', par_para={'num_clients':100}, task_path='./mnist_gen_by_para2')
    """
    if type(benchmark) is str: benchmark = importlib.import_module(benchmark)
    if not hasattr(benchmark, '__path__'): raise RuntimeError("benchmark should be a package or the path of a package")
    if Partitioner is not None:
        if type(Partitioner) is str:
            if Partitioner in globals().keys(): Partitioner = eval(Partitioner)
            else: Partitioner = getattr(flgo.benchmark.toolkits.partition, Partitioner)
        partitioner = Partitioner(**par_para)
    else: partitioner = None
    if rawdata_path!='': bmk_para['rawdata_path']=rawdata_path
    bmk_core = benchmark.core
    task_generator = getattr(bmk_core, 'TaskGenerator')(**bmk_para)
    if partitioner is not None:
        task_generator.register_partitioner(partitioner)
        partitioner.register_generator(task_generator)
    task_generator.generate()
    # save the generated federated benchmark
    # initialize task pipe
    if task_path=='': task_path = os.path.join('.', task_generator.task_name)
    task_pipe = getattr(bmk_core, 'TaskPipe')(task_path)
    # check if task already exists
    if task_pipe.task_exists():
        raise FileExistsError('Task {} already exists.'.format(task_path))
    try:
        # create task architecture
        task_pipe.create_task_architecture()
        # save meta infomation
        task_pipe.save_info(task_generator)
        # save task
        task_pipe.save_task(task_generator)
        print('Task {} has been successfully generated.'.format(task_generator.task_name))
    except Exception as e:
        print(e)
        task_pipe.remove_task()
        print("Failed to saving splited dataset.")
    # save visualization
    try:
        visualize_func = getattr(benchmark,'visualize')
        visualize_func(task_generator, partitioner)
        task_pipe.save_figure()
    except:
        pass

def gen_task_by_config(config={}, task_path:str='', rawdata_path:str='', seed:int=0):
    r"""
    Generate a federated task that is specified by the benchmark information and the partition information, where the generated task will be stored in the task_path and the raw data will be downloaded into the rawdata_path.
    :param
        config (dict || str): configuration is either a dict contains parameters or a filename of a .yml file
        task_path (str): where the generated task will be stored
        rawdata_path (str): where the raw data will be downloaded\stored
        seed (int): the random seed used to generate the task
    :return

    Example:
        >>> import flgo
        >>> config = {'benchmark':{'name':'flgo.benchmark.mnist_classification'}, 'partitioner':{'name':'IIDParitioner', 'para':{'num_clients':100}}}
        >>> flgo.gen_task(config, './my_mnist_iid')
        >>> # The task will be stored as `my_mnist_iid` in the current working dictionary
    """
    # setup random seed
    random.seed(3 + seed)
    np.random.seed(97 + seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    # load configuration
    gen_option = load_configuration(config)
    if 'para' not in gen_option['benchmark'].keys(): gen_option['benchmark']['para'] = {}
    if 'partitioner' in gen_option.keys():
        # update parameters of partitioner
        if 'para' not in gen_option['partitioner'].keys():
            gen_option['partitioner']['para'] = {}
        else:
            if 'name' not in gen_option['partitioner'].keys():
                gen_option['benchmark']['para'].update(gen_option['partitioner']['para'])
    # init generator
    if rawdata_path!='': gen_option['benchmark']['para']['rawdata_path']=rawdata_path
    if type(gen_option['benchmark']['name']) is str:
        bmk_core = importlib.import_module('.'.join([gen_option['benchmark']['name'], 'core']))
    elif hasattr(gen_option['benchmark']['name'], '__path__'):
        bmk_core = gen_option['benchmark']['name'].core
    else:
        raise RuntimeError("The value of parameter config['benchmark']['name'] should be either a string or a python package.")
    task_generator = getattr(bmk_core, 'TaskGenerator')(**gen_option['benchmark']['para'])
    # create partitioner for generator if specified
    if 'partitioner' in gen_option.keys() and 'name' in gen_option['partitioner'].keys():
        Partitioner = gen_option['partitioner']['name']
        if type(Partitioner) is str:
            if Partitioner in globals().keys(): Partitioner = eval(Partitioner)
            else: Partitioner = getattr(flgo.benchmark.toolkits.partition, Partitioner)
        partitioner = Partitioner(**gen_option['partitioner']['para'])
        task_generator.register_partitioner(partitioner)
        partitioner.register_generator(task_generator)
    else:
        partitioner = None
    # generate federated task
    task_generator.generate()
    # save the generated federated benchmark
    # initialize task pipe
    if task_path=='': task_path = os.path.join('.', task_generator.task_name)
    task_pipe = getattr(bmk_core, 'TaskPipe')(task_path)
    # check if task already exists
    if task_pipe.task_exists():
        raise FileExistsError('Task {} already exists.'.format(task_path))
    try:
        # create task architecture
        task_pipe.create_task_architecture()
        # save meta infomation
        task_pipe.save_info(task_generator)
        # save task
        task_pipe.save_task(task_generator)
        print('Task {} has been successfully generated.'.format(task_generator.task_name))
    except Exception as e:
        print(e)
        task_pipe.remove_task()
        print("Failed to saving splited dataset.")
    # save visualization
    try:
        visualize_func = getattr(importlib.import_module(gen_option['benchmark']['name']),'visualize')
        visualize_func(task_generator, partitioner)
        task_pipe.save_figure()
    except:
        pass

def init(task: str, algorithm, option = {}, model=None, Logger: flgo.experiment.logger.BasicLogger = flgo.experiment.logger.simple_logger.SimpleLogger, Simulator: BasicSimulator=flgo.system_simulator.DefaultSimulator, scene='horizontal'):
    r"""
    Initialize a runner in FLGo, which is to optimize a model on a specific task (i.e. IID-mnist-of-100-clients) by the selected federated algorithm.
    :param
        task (str): the dictionary of the federated task
        algorithm (module || class): the algorithm will be used to optimize the model in federated manner, which must contain pre-defined attributions (e.g. algorithm.Server and algorithm.Client for horizontal federated learning)
        option (dict || str): the configurations of training, environment, algorithm, logger and simulator
        model (module || class): the model module that contains two methods: model.init_local_module(object) and model.init_global_module(object)
        Logger (class): the class of the logger inherited from flgo.experiment.logger.BasicLogger
        Simulator (class): the class of the simulator inherited from flgo.system_simulator.BasicSimulator
        scene (str): 'horizontal' or 'vertical' in current version of FLGo
    :return
        runner: the object instance that has the method runner.run()

    Example:
        >>> import flgo
        >>> from flgo.algorithm import fedavg
        >>> from flgo.experiment.logger.simple_logger import SimpleLogger
        >>> # create task 'mnist_iid' by flgo.gen_task('gen_config.yml', 'mnist_iid') if there exists no such task
        >>> if os.path.exists('mnist_iid'): flgo.gen_task({'benchmark':{'name':'flgo.benchmark.mnist_classification'}, 'partitioner':{'name':'IIDPartitioner','para':{'num_clients':100}}}, 'mnist_iid')
        >>> # create runner
        >>> fedavg_runner = flgo.init('mnist_iid', algorithm=fedavg, option = {'num_rounds':20, 'gpu':[0], 'learning_rate':0.1})
        >>> fedavg_runner.run()
        ... # the training will start after runner.run() was called, and the running-time results will be recorded by Logger into the task dictionary
    """

    # init option
    option = load_configuration(option)
    default_option = read_option_from_command()
    for op_key in option:
        if op_key in default_option.keys():
            op_type = type(default_option[op_key])
            if op_type == type(option[op_key]):
                default_option[op_key] = option[op_key]
            else:
                if op_type is list:
                    default_option[op_key]=list(option[op_key]) if hasattr(option[op_key], '__iter__') else [option[op_key]]
                elif op_type is tuple:
                    default_option[op_key] = tuple(option[op_key]) if hasattr(option[op_key], '__iter__') else (option[op_key])
                else:
                    default_option[op_key] = op_type(option[op_key])
    option = default_option
    setup_seed(seed=option['seed'])
    option['task'] = task
    option['algorithm'] = (algorithm.__name__).split('.')[-1]

    # init task info
    if not os.path.exists(task):
        raise FileExistsError("Fedtask '{}' doesn't exist. Please generate the specified task by flgo.gen_task().")
    with open(os.path.join(task, 'info'), 'r') as inf:
        task_info = json.load(inf)
    benchmark = task_info['benchmark']
    if model== None: model = getattr(importlib.import_module(benchmark), 'default_model')
    option['model'] = (model.__name__).split('.')[-1]

    try:
        # init multiprocess
        torch.multiprocessing.set_start_method('spawn')
        torch.multiprocessing.set_sharing_strategy('file_system')
    except:
        pass

    # create global variable
    gv = GlobalVariable()
    # init logger
    gv.logger = Logger(task=task, option=option, name=str(id(gv))+str(Logger), level=option['log_level'])

    # init device
    gv.dev_list = [torch.device('cpu')] if (option['gpu'] is None or len(option['gpu'])==0) else [torch.device('cuda:{}'.format(gpu_id)) for gpu_id in option['gpu']]
    gv.logger.info('Initializing devices: '+','.join([str(dev) for dev in gv.dev_list])+' will be used for this running.')
    # init task
    core_module = '.'.join([benchmark, 'core'])
    gv.TaskPipe = getattr(importlib.import_module(core_module), 'TaskPipe')
    task_pipe = gv.TaskPipe(task)
    gv.TaskCalculator = getattr(importlib.import_module(core_module), 'TaskCalculator')
    task_data = task_pipe.load_data(option)

    # init objects
    obj_class = [c for c in dir(algorithm) if not c.startswith('__')]
    tmp = []
    for c in obj_class:
        try:
            C = getattr(algorithm, c)
            setattr(C, 'gv', gv)
            tmp.append(c)
        except:
            continue
    objects = task_pipe.generate_objects(option, algorithm, scene=scene)
    task_pipe.distribute(task_data, objects)

    for c in tmp:
        try:
            C = getattr(algorithm, c)
            delattr(C, 'gv')
        except:
            continue

    # init model
    if hasattr(model, 'init_local_module'):
        for object in objects:
            model.init_local_module(object)
    if hasattr(model, 'init_global_module'):
        for object in objects:
            model.init_global_module(object)

    # init communicator
    gv.communicator = flgo.VirtualCommunicator(objects)

    for ob in objects: ob.initialize()

    # init virtual system environment
    gv.logger.info('Use `{}` as the system simulator'.format(str(Simulator)))
    flgo.system_simulator.base.random_seed_gen = flgo.system_simulator.base.seed_generator(option['seed'])
    gv.clock = flgo.system_simulator.base.ElemClock()
    gv.simulator = Simulator(objects, option)
    gv.clock.register_simulator(simulator=gv.simulator)

    gv.logger.register_variable(coordinator=objects[0], participants=objects[1:], option=option, clock=gv.clock, scene=scene, objects = objects)
    gv.logger.initialize()
    gv.logger.info('Ready to start.')

    # register global variables for objects
    for ob in objects:
        ob.gv = gv
    gv.simulator.gv = gv
    gv.clock.gv = gv
    gv.logger.gv = gv
    return objects[0]

def _call_by_process(task, algorithm_name,  opt, model_name, Logger, Simulator, scene, send_end):
    pid = os.getpid()
    sys.stdout = open(os.devnull, 'w')
    if model_name is None: model = None
    else:
        try:
            model = importlib.import_module(model_name)
        except:
            model = model_name
    try:
        algorithm = importlib.import_module(algorithm_name)
    except:
        algorithm = algorithm_name
    try:
        runner = flgo.init(task, algorithm, model=model, option=opt, Logger=Logger, Simulator=Simulator, scene=scene)
        runner.run()
        res = (os.path.join(runner.gv.logger.get_output_path(), runner.gv.logger.get_output_name()), pid)
        send_end.send(res)
    except Exception as e:
        s = 'Process {} exits with error:" {}". '.format(pid, str(e))
        res = (opt, s, pid)
        send_end.send(res)

def get_available_device(device_ids):
    # dev_handlers = [pynvml.nvmlDeviceGetHandleByIndex(dev_id) for dev_id in device_ids]
    return random.choice(device_ids)

def tune(task: str, algorithm, option: dict = {}, model=None, Logger: flgo.experiment.logger.BasicLogger = flgo.experiment.logger.tune_logger.TuneLogger, Simulator: BasicSimulator=flgo.system_simulator.DefaultSimulator, scene='horizontal', scheduler=None):
    """
        Tune hyper-parameters for one task and one algorithm in parallel.
        :param
            task (str): the dictionary of the federated task
            algorithm (module || class): the algorithm will be used to optimize the model in federated manner, which must contain pre-defined attributions (e.g. algorithm.Server and algorithm.Client for horizontal federated learning)
            option (dict): the dict whose values should be of type list to construct the combinations
            model (module || class): the model module that contains two methods: model.init_local_module(object) and model.init_global_module(object)
            Logger (class): the class of the logger inherited from flgo.experiment.logger.BasicLogger
            Simulator (class): the class of the simulator inherited from flgo.system_simulator.BasicSimulator
            scene (str): 'horizontal' or 'vertical' in current version of FLGo
            scheduler (instance of flgo.experiment.device_scheduler.BasicScheduler): GPU scheduler that schedules GPU by checking their availability
        """
    # generate combinations of hyper-parameters
    if 'gpu' in option.keys():
        device_ids = option['gpu']
        option.pop('gpu')
        if not isinstance(device_ids, Iterable): device_ids = [device_ids]
    else:
        device_ids = [-1]
    keys = list(option.keys())
    for k in keys: option[k] = [option[k]] if (not isinstance(option[k], Iterable) or isinstance(option[k], str)) else option[k]
    para_combs = [para_comb for para_comb in itertools.product(*(option[k] for k in keys))]
    options = [{k:v for k,v in zip(keys, paras)} for paras in para_combs]
    for op in options:op['log_file'] = True
    if scheduler is None:
        scheduler = flgo.experiment.device_scheduler.BasicScheduler(device_ids)
    outputs = run_in_parallel(task, algorithm, options,model, devices=device_ids, Logger=Logger, Simulator=Simulator, scene=scene, scheduler=scheduler)
    optimal_idx = int(np.argmin([min(output['valid_loss']) for output in outputs]))
    optimal_para = options[optimal_idx]
    print("The optimal combination of hyper-parameters is:")
    print('-----------------------------------------------')
    for k,v in optimal_para.items():
        if k=='gpu': continue
        print("{}\t|{}".format(k,v))
    print('-----------------------------------------------')
    op_round = np.argmin(outputs[optimal_idx]['valid_loss'])
    if 'eval_interval' in option.keys(): op_round = option['eval_interval']*op_round
    print('The minimal validation loss occurs at the round {}'.format(op_round))

def run_in_parallel(task: str, algorithm, options:list = [], model=None, devices = [], Logger:flgo.experiment.logger.BasicLogger = flgo.experiment.logger.simple_logger.SimpleLogger, Simulator=flgo.system_simulator.DefaultSimulator, scene='horizontal', scheduler = None):
    """
    Run different groups of hyper-parameters for one task and one algorithm in parallel.
    :param
        task (str): the dictionary of the federated task
        algorithm (module || class): the algorithm will be used to optimize the model in federated manner, which must contain pre-defined attributions (e.g. algorithm.Server and algorithm.Client for horizontal federated learning)
        options (list): the configurations of different groups of hyper-parameters
        model (module || class): the model module that contains two methods: model.init_local_module(object) and model.init_global_module(object)
        devices (list): the list of IDs of devices
        Logger (class): the class of the logger inherited from flgo.experiment.logger.BasicLogger
        Simulator (class): the class of the simulator inherited from flgo.system_simulator.BasicSimulator
        scene (str): 'horizontal' or 'vertical' in current version of FLGo
        scheduler (instance of flgo.experiment.device_scheduler.BasicScheduler): GPU scheduler that schedules GPU by checking their availability
    :return
        the returns of _call_by_process
    """
    try:
        # init multiprocess
        torch.multiprocessing.set_start_method('spawn', force=True)
        torch.multiprocessing.set_sharing_strategy('file_system')
    except:
        pass
    if model is None:
        model_name = None
    else:
        if not hasattr(model, '__module__') and hasattr(model, '__name__'):
            model_name = model.__name__
        else:
            model_name = model
    algorithm_name = algorithm.__name__ if (not hasattr(algorithm, '__module__') and hasattr(algorithm, '__name__')) else algorithm
    option_state = {oid:{'p':None, 'completed':False, 'output':None, 'option_in_queue':False, 'recv':None, } for oid in range(len(options))}
    if scheduler is None: scheduler = flgo.experiment.device_scheduler.BasicScheduler(devices)
    while True:
        for oid in range(len(options)):
            opt = options[oid]
            if option_state[oid]['p'] is None:
                if not option_state[oid]['completed']:
                    available_device = scheduler.get_available_device(opt)
                    if available_device is None: continue
                    else:
                        opt['gpu'] = available_device
                        recv_end, send_end = multiprocessing.Pipe(False)
                        option_state[oid]['p'] = multiprocessing.Process(target=_call_by_process, args=(task, algorithm_name, opt, model_name, Logger, Simulator, scene, send_end))
                        option_state[oid]['recv'] = recv_end
                        option_state[oid]['p'].start()
                        scheduler.add_process(option_state[oid]['p'].pid)
                        print('Process {} was created for args {}'.format(option_state[oid]['p'].pid,(task, algorithm_name, opt, model_name, Logger, Simulator, scene)))
            else:
                if option_state[oid]['p'].exitcode is not None:
                    tmp = option_state[oid]['recv'].recv()
                    scheduler.remove_process(tmp[-1])
                    try:
                        option_state[oid]['p'].terminate()
                    except:
                        pass
                    option_state[oid]['p'] = None
                    if len(tmp)==2:
                        option_state[oid]['completed'] = True
                        option_state[oid]['output'] = tmp[0]
                    else:
                        print(tmp[1])
        if all([v['completed'] for v in option_state.values()]):break
        time.sleep(1)
    res = []
    for oid in range(len(options)):
        rec_path = option_state[oid]['output']
        with open(rec_path, 'r') as inf:
            s_inf = inf.read()
            rec = json.loads(s_inf)
        res.append(rec)
    return res

def multi_init_and_run(runner_args:list, devices = [], scheduler=None):
    r"""
    Create multiple runners and run in parallel
    param:
        runner_args (list): each element in runner_args should be either a dict or a tuple or parameters
        devices (list): a list of gpu id
        scheduler (class flgo.experiment.device_scheduler.BasicScheduler): GPU scheduler
    return:
        a list of output results of runners
    Example:
        >>> from flgo.algorithm import fedavg, fedprox, scaffold
        >>> # create task 'mnist_iid' by flgo.gen_task if there exists no such task
        >>> task='./mnist_iid'
        >>> if os.path.exists(task): flgo.gen_task({'benchmark':{'name':'flgo.benchmark.mnist_classification'}, 'partitioner':{'name':'IIDPartitioner','para':{'num_clients':100}}}, task)
        >>> algos = [fedavg, fedprox, scaffold]
        >>> flgo.multi_init_and_run([{'task':task, 'algorithm':algo} for algo in algos], devices=[0])
    """
    if len(runner_args)==0:return
    args = []
    if type(runner_args[0]) is dict:
        for a in runner_args:
            tmp = collections.defaultdict(lambda:None, a)
            if tmp['task'] is None or tmp['algorithm'] is None:
                raise RuntimeError("keyword 'task' or 'algorithm' is of NoneType")
            algorithm = tmp['algorithm']
            tmp['algorithm'] = algorithm.__name__ if (not hasattr(algorithm, '__module__') and hasattr(algorithm, '__name__')) else algorithm
            if tmp['option'] is None:
                tmp['option'] = default_option_dict
            else:
                option = tmp['option']
                default_option = read_option_from_command()
                for op_key in option:
                    if op_key in default_option.keys():
                        op_type = type(default_option[op_key])
                        if op_type == type(option[op_key]):
                            default_option[op_key] = option[op_key]
                        else:
                            if op_type is list:
                                default_option[op_key] = list(option[op_key]) if hasattr(option[op_key],
                                                                                         '__iter__') else [
                                    option[op_key]]
                            elif op_type is tuple:
                                default_option[op_key] = tuple(option[op_key]) if hasattr(option[op_key],
                                                                                          '__iter__') else (
                                option[op_key])
                            else:
                                default_option[op_key] = op_type(option[op_key])
                tmp['option'] = default_option
            if tmp['model'] is None:
                model_name = None
            else:
                if not hasattr(tmp['model'], '__module__') and hasattr(tmp['model'], '__name__'):
                    model_name = tmp['model'].__name__
                else:
                    model_name = tmp['model']
            tmp['model'] = model_name
            if tmp['Logger'] is None:
                tmp['Logger'] = flgo.experiment.logger.simple_logger.SimpleLogger
            algorithm_name = tmp['algorithm'].__name__ if (not hasattr(tmp['algorithm'], '__module__') and hasattr(tmp['algorithm'], '__name__')) else tmp['algorithm']
            if tmp['Simulator'] is None:
                tmp['Simulator'] = flgo.system_simulator.DefaultSimulator
            if tmp['scene'] is None:
                tmp['scene'] = 'horizontal'
            args.append(list(tmp.values()))
    elif type(runner_args[0]) is tuple or type(runner_args[0]) is list:
        for a in runner_args:
            if len(a)<2: raise RuntimeError('the args of runner should at least contain task and algorithm.')
            default_args = [None, None, default_option_dict, None, flgo.experiment.logger.simple_logger.SimpleLogger, flgo.system_simulator.DefaultSimulator, 'horizontal']
            for aid in range(len(a)):
                if aid==0:
                    default_args[aid] = a[aid]
                if aid==1:
                    algorithm = a[aid]
                    algorithm_name = algorithm.__name__ if (not hasattr(algorithm, '__module__') and hasattr(algorithm, '__name__')) else algorithm
                    default_args[aid] = a[aid]
                elif aid==2:
                    option = a[aid]
                    default_option = read_option_from_command()
                    for op_key in option:
                        if op_key in default_option.keys():
                            op_type = type(default_option[op_key])
                            if op_type == type(option[op_key]):
                                default_option[op_key] = option[op_key]
                            else:
                                if op_type is list:
                                    default_option[op_key] = list(option[op_key]) if hasattr(option[op_key],
                                                                                             '__iter__') else [
                                        option[op_key]]
                                elif op_type is tuple:
                                    default_option[op_key] = tuple(option[op_key]) if hasattr(option[op_key],
                                                                                              '__iter__') else (
                                        option[op_key])
                                else:
                                    default_option[op_key] = op_type(option[op_key])
                    default_args[aid] = default_option
                elif aid==3:
                    model = a[aid]
                    if model is None:
                        model_name = None
                    else:
                        if not hasattr(model, '__module__') and hasattr(model, '__name__'):
                            model_name = model.__name__
                        else:
                            model_name = model
                    default_args[aid] = model_name
                else:
                    default_args[aid] = a[aid]

    runner_state = {rid: {'p': None, 'completed': False, 'output': None, 'runner_in_queue': False, 'recv': None, } for
                    rid in range(len(args))}
    if scheduler is None: scheduler = flgo.experiment.device_scheduler.BasicScheduler(devices)
    while True:
        for rid in range(len(args)):
            current_arg = args[rid]
            if runner_state[rid]['p'] is None:
                if not runner_state[rid]['completed']:
                    available_device = scheduler.get_available_device(current_arg)
                    if available_device is None:
                        continue
                    else:
                        list_current_arg = copy.deepcopy(current_arg)
                        list_current_arg[2]['gpu'] = available_device
                        recv_end, send_end = multiprocessing.Pipe(False)
                        list_current_arg.append(send_end)
                        runner_state[rid]['p'] = multiprocessing.Process(target=_call_by_process, args=tuple(list_current_arg))
                        runner_state[rid]['recv'] = recv_end
                        runner_state[rid]['p'].start()
                        scheduler.add_process(runner_state[rid]['p'].pid)
                        print('Process {} was created for args {}'.format(runner_state[rid]['p'].pid,current_arg))
            else:
                if runner_state[rid]['p'].exitcode is not None:
                    tmp = runner_state[rid]['recv'].recv()
                    scheduler.remove_process(tmp[-1])
                    try:
                        runner_state[rid]['p'].terminate()
                    except:
                        pass
                    runner_state[rid]['p'] = None
                    if len(tmp) == 2:
                        runner_state[rid]['completed'] = True
                        runner_state[rid]['output'] = tmp[0]
                    else:
                        print(tmp[1])
        if all([v['completed'] for v in runner_state.values()]): break
        time.sleep(1)
    res = []
    for rid in range(len(runner_state)):
        rec_path = runner_state[rid]['output']
        with open(rec_path, 'r') as inf:
            s_inf = inf.read()
            rec = json.loads(s_inf)
        res.append(rec)
    return res





import torch
from torch_geometric.nn import GCNConv
from flgo.utils.fmodule import FModule

class Model(FModule):
    def __init__(self):
        super(Model, self).__init__()
        self.conv1 = GCNConv(1433, 128)
        self.conv2 = GCNConv(128, 64)

    def encode(self, x, edge_index):
        x = self.conv1(x, edge_index)
        x = x.relu()
        return self.conv2(x, edge_index)

    def decode(self, z, edge_label_index):
        return (z[edge_label_index[0]] * z[edge_label_index[1]]).sum(dim=-1)

    def decode_all(self, z):
        prob_adj = z @ z.t()
        return (prob_adj > 0).nonzero(as_tuple=False).t()

def init_local_module(object):
    pass

def init_global_module(object):
    if 'Server' in object.__class__.__name__:
        object.model = Model().to(object.device)


import math
import shutil
import urllib.request
import random
import zipfile
import flgo.benchmark
import os.path
import json
import numpy as np
from sklearn.model_selection import train_test_split
from torchvision.datasets import MNIST, utils
from PIL import Image
import os.path
import torch
import torchvision
from tqdm import tqdm

from flgo.benchmark.toolkits.cv.horizontal.image_classification import BuiltinClassGenerator, BuiltinClassPipe, GeneralCalculator

def download_from_url(url= None, filepath = '.'):
    """Download dataset from url to filepath."""
    if url:urllib.request.urlretrieve(url, filepath)
    return filepath

def extract_from_zip(src_path, target_path):
    """Unzip the .zip file (src_path) to target_path"""
    f = zipfile.ZipFile(src_path)
    f.extractall(target_path)
    targets = f.namelist()
    f.close()
    return [os.path.join(target_path, tar) for tar in targets]

class FEMNIST(MNIST):
    """
    This dataset is derived from the Leaf repository
    (https://github.com/TalwalkarLab/leaf) pre-processing of the Extended MNIST
    dataset, grouping examples by writer. Details about Leaf were published in
    "LEAF: A Benchmark for Federated Settings" https://arxiv.org/abs/1812.01097.
    """

    def __init__(self, root, train=True, transform=None, target_transform=None,
                 download=False):
        super(MNIST, self).__init__(root, transform=transform,
                                    target_transform=target_transform)
        self.train = train
        if self.train:
            data_file = self.training_file
        else:
            data_file = self.test_file
        if not os.path.exists(os.path.join(self.processed_folder, data_file)):
            self.download_and_process()
        self.data, self.targets, self.user_index = torch.load(os.path.join(self.processed_folder, data_file))

    def __getitem__(self, index):
        img, target = self.data[index], int(self.targets[index])
        img = Image.fromarray(img.numpy(), mode='F')
        if self.transform is not None:
            img = self.transform(img)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return img, target

    def download_and_process(self):
        """
        Download the raw data and process it and save it in Torch format
        Modified from https://github.com/alibaba/FederatedScope/blob/master/federatedscope/cv/dataset/leaf_cv.py
        """
        if not os.path.exists(os.path.join(self.raw_folder, 'all_data.json')):
            """Download the FEMNIST data if it doesn't exist in processed_folder already."""

            os.makedirs(self.raw_folder, exist_ok=True)
            os.makedirs(self.processed_folder, exist_ok=True)

            # Download to `self.raw_dir`.
            url = 'https://federatedscope.oss-cn-beijing.aliyuncs.com'
            name = 'femnist_all_data.zip'
            src_path = download_from_url(f'{url}/{name}', os.path.join(self.raw_folder, 'tmp'))
            tar_paths = extract_from_zip(src_path, self.raw_folder)
            all_data = {
                'users': [],
                'num_samples': [],
                'user_data': {}
            }
            for i in range(1, len(tar_paths)):
                with open(tar_paths[i], 'r') as f:
                    raw_data = json.load(f)
                    all_data['users'].extend(raw_data['users'])
                    all_data['num_samples'].extend(raw_data['num_samples'])
                    all_data['user_data'].update(raw_data['user_data'])
            with open(os.path.join(self.raw_folder, 'all_data.json'), 'w') as f:
                json.dump(all_data, f)
            os.remove(src_path)
            shutil.rmtree(tar_paths[0])
        else:
            with open(os.path.join(self.raw_folder, 'all_data.json'), 'r') as f:
                all_data = json.load(f)
        """Process Data"""
        Xs = []
        Ys = []
        sample_ids = []
        idx = 0
        for writer, v in all_data['user_data'].items():
            data, targets = v['x'], v['y']
            Xs.extend(data)
            Ys.extend(targets)
            sample_ids.extend([idx] * len(data))
            idx += 1
        Xs = torch.tensor(np.stack(Xs))
        Ys = torch.LongTensor(np.stack(Ys))
        sample_ids = torch.tensor(np.stack(sample_ids))
        num_samples = sample_ids.shape[0]
        s1 = int(num_samples * 0.9)
        s2 = num_samples - s1
        train_ids, test_ids = torch.utils.data.random_split(sample_ids, [s1, s2])
        train_indices = train_ids.indices
        test_indices = test_ids.indices
        train_data, train_targets, train_sample_id = Xs[train_indices], Ys[train_indices], sample_ids[train_indices]
        test_data, test_targets, test_sample_id = Xs[test_indices], Ys[test_indices], sample_ids[test_indices]
        torch.save((train_data, train_targets, train_sample_id), os.path.join(self.processed_folder, "training.pt"))
        torch.save((test_data, test_targets, test_sample_id), os.path.join(self.processed_folder, "test.pt"))

TaskCalculator = GeneralCalculator

class TaskGenerator(BuiltinClassGenerator):
    def __init__(self, rawdata_path=os.path.join(flgo.benchmark.path,'RAW_DATA', 'MNIST')):
        super(TaskGenerator, self).__init__('femnist_classification', rawdata_path, FEMNIST, torchvision.transforms.Compose([torchvision.transforms.ToTensor(),]))

class TaskPipe(BuiltinClassPipe):
    def __init__(self, task_name):
        super(TaskPipe, self).__init__(task_name, FEMNIST, torchvision.transforms.Compose([torchvision.transforms.ToTensor(),]))


import torchvision
from flgo.benchmark.toolkits.cv.horizontal.image_classification import BuiltinClassGenerator, BuiltinClassPipe, GeneralCalculator
import flgo.benchmark
import os.path
builtin_class = torchvision.datasets.CIFAR100
transforms = torchvision.transforms.Compose([torchvision.transforms.ToTensor(), torchvision.transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010))])
TaskCalculator = GeneralCalculator

class TaskGenerator(BuiltinClassGenerator):
    def __init__(self, rawdata_path=os.path.join(flgo.benchmark.path,'RAW_DATA', 'CIFAR10')):
        super(TaskGenerator, self).__init__('cifar100_classification', rawdata_path, builtin_class, transforms)

class TaskPipe(BuiltinClassPipe):
    def __init__(self, task_name):
        super(TaskPipe, self).__init__(task_name, builtin_class, transforms)


from flgo.algorithm.fedbase import BasicServer, BasicClient
import copy
from flgo.utils import fmodule
import torch

class Server(BasicServer):
    def initialize(self, *args, **kwargs):
        self.init_algo_para({'eta': 1.0})
        self.cg = self.model.zeros_like()
        self.sample_option = 'uniform'

    def pack(self, client_id, *args, **kwargs):
        return {
            "model": copy.deepcopy(self.model),
            "cg": self.cg,
        }

    def iterate(self):
        # sample clients
        self.selected_clients = self.sample()
        # local training
        res = self.communicate(self.selected_clients)
        dys, dcs = res['dy'], res['dc']
        # aggregate
        self.model, self.cg = self.aggregate(dys, dcs)
        return

    def aggregate(self, dys, dcs):
        # x <-- x + eta_g * dx = x + eta_g * average(dys)
        # c <-- c + |S|/N * dc = c + 1/N * sum(dcs)
        new_model = self.model + self.eta * fmodule._model_average(dys)
        new_c = self.cg + fmodule._model_sum(dcs)/self.num_clients
        return new_model, new_c


class Client(BasicClient):
    def initialize(self, *args, **kwargs):
        self.c = None

    @fmodule.with_multi_gpus
    def train(self, model, cg):
        """
        The codes of Algorithm 1 that updates the control variate
          12:  ci+ <-- ci - c + 1 / K / eta_l * (x - yi)
          13:  communicate (dy, dc) <-- (yi - x, ci+ - ci)
          14:  ci <-- ci+
        Our implementation for efficiency
          dy = yi - x
          dc <-- ci+ - ci = -1/K/eta_l * (yi - x) - c = -1 / K /eta_l *dy - c
          ci <-- ci+ = ci + dc
          communicate (dy, dc)
        """
        model.train()
        # global parameters
        src_model = copy.deepcopy(model)
        src_model.freeze_grad()
        cg.freeze_grad()
        # if self.c is None: self.c = copy.deepcopy(cg)
        if self.c is None: self.c = cg.zeros_like()
        self.c.freeze_grad()
        optimizer = self.calculator.get_optimizer(model, lr=self.learning_rate, weight_decay=self.weight_decay,
                                                  momentum=self.momentum)
        for iter in range(self.num_steps):
            batch_data = self.get_batch_data()
            model.zero_grad()
            loss = self.calculator.compute_loss(model, batch_data)['loss']
            loss.backward()
            # y_i <-- y_i - eta_l ( g_i(y_i)-c_i+c )  =>  g_i(y_i)' <-- g_i(y_i)-c_i+c
            for pm, pcg, pc in zip(model.parameters(), cg.parameters(), self.c.parameters()):
                pm.grad = pm.grad - pc + pcg
            optimizer.step()
        dy = model - src_model
        dc = -dy/(self.num_steps * self.learning_rate) - cg
        self.c = self.c + dc
        return dy, dc

    def pack(self, dy, dc):
        return {
            "dy": dy,
            "dc": dc,
        }

    def unpack(self, received_pkg):
        # unpack the received package
        return received_pkg['model'], received_pkg['cg']

    def reply(self, svr_pkg):
        model, c_g = self.unpack(svr_pkg)
        dy, dc = self.train(model, c_g)
        cpkg = self.pack(dy, dc)
        return cpkg


import torch
import argparse
import torch.nn as nn
import torch.nn.functional as F
import torch_geometric.transforms as T
import torch_geometric.nn as pyg_nn
from flgo.utils.fmodule import FModule
from torch.nn import Sequential, Linear, ReLU

class Model(FModule):
    def __init__(self, input_dim=3, hidden_dim=20, output_dim=6):
        super(Model, self).__init__()
        self.num_layers = 3
        self.pre_mp = nn.Sequential(
            nn.Linear(input_dim, hidden_dim))

        self.convs = nn.ModuleList()
        self.nn1 = Sequential(Linear(hidden_dim, hidden_dim), ReLU(), Linear(hidden_dim, hidden_dim))
        self.convs.append(pyg_nn.GINConv(self.nn1))
        for l in range(self.num_layers - 1):
            self.nnk = Sequential(Linear(hidden_dim, hidden_dim), ReLU(), Linear(hidden_dim, hidden_dim))
            self.convs.append(pyg_nn.GINConv(self.nnk))

        self.post_mp = nn.Sequential(
            nn.Linear(hidden_dim, hidden_dim), ReLU(), nn.Linear(hidden_dim, output_dim))

    def forward(self, data):
        x, edge_index, batch = data.x, data.edge_index, data.batch
        x = self.pre_mp(x)
        for i in range(len(self.convs)):
            x = self.convs[i](x, edge_index)
            x = F.relu(x)
        x = pyg_nn.global_add_pool(x, batch)
        x = self.post_mp(x)
        x = F.log_softmax(x, dim=1)
        return x

def init_local_module(object):
    pass

def init_global_module(object):
    if 'Server' in object.__class__.__name__:
        object.model = Model().to(object.device)


r"""
This module is to analyze the training results saved by Logger. To use this module,
a analysis plan must be designed as a dict that contains three parts:
    Selector: select the records according to the task, algorithm and options of the task
    Painter: draw graphic of the selected records
    Table: output some statistic of the selected records on the console

The basic usage is to build a plan dict and pass it to flgo.experiment.analyzer
>>> # plan = {'Selector':..., 'Painter':..., 'Table':...,}
>>> flgo.experiment.analyzer.show(plan)

The following three examples show how to build a customized plan:

Example 1: How to define a Selector?
    {'Selector': {
        'task': task_path,                # all the analysis will be conducted on a single task
        'header': ['fedavg'],             # only the records where the names of algorithms are in `header` will be selected
         'filter': {'LR':'<0.1'}          # only the records whose options satisfy the conditions in `filter` will be selected
        'legend_with': ['LR', 'B', 'E']   # all the graphic will show the legends of records according to `legend_with`
    }, ...}

Example 2: How to define a Painter?
        Each `Painter` is a dict of different types of graphic (e.g. Curve, Bar and Scatter). In each types of graphic,
        the value is a list of figures, where each figure is defined by a dict like {'args':{...}, 'obj_option':{}, 'fig_option':{...}}
    {...,
    'Painter':{
            'Curve':[
                {'args':{'x':'communication_round', 'y':'valid_loss'}, },
                {...}
            ]
        },
    ...,
    }

Example 3: How to define a Table?
    {...,
    'Table':{
            'min_value':[
                {'x':'valid_loss'},
                ...
                ]
        }
    }

A standard analysis plan usually consists of the above three parts, and `Painter` and `Table` are both optional
"""
import argparse
import math
import random
import numpy as np
import matplotlib.pyplot as plt
import yaml
import os
import collections
import copy
import matplotlib as mpl
import prettytable as pt
import json
from flgo.utils.fflow import load_configuration

def option2filter(option: dict):
    valid_keys = {
        'learning_rate': 'LR',
        'batch_size': 'B',
        'num_rounds': 'R',
        'num_epochs': 'E',
        'num_steps': 'K',
        'proportion': 'P',
        'model':'M',
        'seed': 'S',
        'learning_rate_decay':'LD',
        'weight_decay': 'WD',
        'availability': 'AVL',
        'connectivity': 'CN',
        'completeness': 'CP',
        'responsiveness': 'RS',
    }

class Record:
    def __init__(self, task, name):
        self.task = task
        self.name = name
        self.rec_path = os.path.join(task, 'record', name)
        with open(self.rec_path, 'r') as inf:
            s_inf = inf.read()
            rec = json.loads(s_inf)
        self.data = rec
        self.set_communication_round()
        self.set_client_id()

    def set_communication_round(self):
        num_rounds = self.data['option']['num_rounds']
        eval_interval = self.data['option']['eval_interval']
        x = [0]
        for round in range(1, num_rounds + 1):
            if eval_interval > 0 and (round == 0 or round % eval_interval == 0):
                x.append(round)
            if self.data['option']['early_stop'] > 0 and 'valid_loss' in self.data.keys() and len(x) >= len(self.data['valid_loss']):
                break
        self.data['communication_round'] = x

    def set_client_id(self):
        with open(os.path.join(self.task, 'info')) as inf:
            num_clients = json.load(inf)['num_clients']
        self.data['client_id'] = [cid for cid in range(int(num_clients))]

    def set_legend(self, legend_with = []):
        if len(legend_with)==0: self.data['label'] = []
        self.data['label'] = [self.name[:self.name.find('_M')]]
        for key in legend_with:
            val = key + self.get_key_from_name(key)
            self.data['label'].append(val)
        self.data['label'] = ' '.join(self.data['label'])

    def get_key_from_name(self, key):
        if key == '': return ''
        value_start = self.name.find('_' + key) + len(key) + 1
        value_end = self.name.find('_', value_start)
        return self.name[value_start:value_end]

    @classmethod
    def group_records(cls, rec_list: list):
        if len(rec_list) == 0: return None
        r = copy.deepcopy(rec_list[0])
        r.datas = [rec.data for rec in rec_list]
        for key in r.data.keys():
            if key == 'option': continue
            if type(r.data[key]) is list:
                ave_data = np.array([np.array(rdata[key]) for rdata in r.datas])
                r.data[key] = ave_data.mean(axis=0)
        return r
############################## Selector ##############################
class Selector:
    def __init__(self, selector_config):
        self.config = selector_config
        self.tasks = [selector_config['task']] if type(selector_config['task']) is not list else selector_config['task']
        self.headers = selector_config['header']
        self.filter = selector_config['filter'] if 'filter' in selector_config.keys() else {}
        self.legend_with = selector_config['legend_with'] if 'legend_with' in selector_config.keys() else []
        self.rec_names = self.scan()
        self.records = self.read_records(self.rec_names)
        try:
            self.group_names, self.grouped_records = self.group_records_by_seed()
        except Exception() as e:
            print(e)

    def scan(self):
        res = {}
        for task in self.tasks:
            path = os.path.join(task, 'record')
            all_records = os.listdir(path)
            tmp = []
            # check headers
            for header in self.headers:
                tmp.extend([f for f in all_records if f.startswith(header) and f.endswith('.json')])
            res[task] = self.filename_filter(tmp, self.filter)
        return res

    def filename_filter(self, fnames, filter):
        if len(filter)==0: return fnames
        for key in filter.keys():
            condition = filter[key]
            res = []
            for f in fnames:
                if f.find('_'+key)==-1: continue
                fv = f[f.find('_' + key) + len(key) + 1:f.find('_', f.find('_' + key) + 1)]
                if type(condition) is list:
                    fv = float(fv) if ('0' <= fv[0] <= '9' or fv[0] == '.' or fv[0] == '-') else fv
                    if fv in condition: res.append(f)
                elif type(condition) is str:
                    con = (fv+condition) if condition[0] in ['<', '>', '='] else (fv+'=='+condition)
                    if eval(con): res.append(f)
                else:
                    if float(fv)==float(condition): res.append(f)
            fnames = res
        return fnames

    def get_key_from_filename(self, filename, key):
        if key == '': return ''
        value_start = filename.find('_' + key) + len(key) + 1
        value_end = filename.find('_', value_start)
        return filename[value_start:value_end]

    def read_records(self, rec_names):
        res = {task: [] for task in rec_names}
        for task in rec_names:
            path = os.path.join(task, 'record')
            files = os.listdir(path)
            for record_name in rec_names[task]:
                if record_name in files:
                    record = Record(task, record_name)
                    record.set_legend(self.legend_with)
                    res[task].append(record)
        return res

    def group_records_by_seed(self):
        group_names = {task:[] for task in self.rec_names}
        grouped_records = {task:[] for task in self.rec_names}
        for task in self.rec_names:
            groups = collections.defaultdict(list)
            for rec in self.records[task]:
                s = rec.name.find('_S')
                g = rec.name[:s] + rec.name[rec.name.find('_', s+1):]
                groups[g].append(rec)
            for g in groups:
                group_names[task].append(g)
                grouped_records[task].append(Record.group_records(groups[g]))
        # aggregate different records within the same group
        return group_names, grouped_records

##############################  Painter ##############################
class PaintObject:
    def __init__(self, rec: dict, args: dict,  obj_option: dict, draw_func: str):
        self.rec = rec
        self.args = args
        self.obj_option = obj_option
        self.draw_func = draw_func
        self.para = (rec.data[v] for v in args.values())
        self.with_legend = True

    def draw(self, ax):
        if 'label' in self.obj_option.keys() or 'label' not in self.rec.data.keys():
            eval('ax.'+str(self.draw_func)+'(*self.para, **self.obj_option)')
        else:
            eval('ax.' + str(self.draw_func) + '(*self.para, **self.obj_option, label=self.rec.data["label"])')
        if self.with_legend: eval('ax.legend()')
        return

class Curve(PaintObject):
    def __init__(self, rec, args,  obj_option):
        super(Curve, self).__init__(rec, args, obj_option, 'plot')

class Bar(PaintObject):
    def __init__(self, rec, args,  obj_option):
        super(Bar, self).__init__(rec, args, obj_option, 'bar')

class Scatter(PaintObject):
    def __init__(self, rec, args,  obj_option):
        super(Scatter, self).__init__(rec, args, obj_option, 'scatter')

class Trace2D(PaintObject):
    def __init__(self, rec, args,  obj_option):
        super(Trace2D, self).__init__(rec, args, obj_option, '')

    def draw(self, ax):
        pass

class GroupCurve(PaintObject):
    def __init__(self, rec, args,  obj_option):
        super(GroupCurve, self).__init__(rec, args, obj_option, '')

    def draw(self, ax):
        x = self.rec.data[self.args['x']]
        ykey = self.args['y']
        mean_y = self.rec.data[ykey]
        min_y = np.min(np.array([d[ykey] for d in self.rec.datas]), axis=0)
        max_y = np.max(np.array([d[ykey] for d in self.rec.datas]), axis=0)
        ax.plot(x, mean_y, label=self.rec.data['label'])
        ax.fill_between(x, max_y, min_y, alpha=0.3)
        ax.legend()

class GroupBar(Bar):
    pass

class Painter:
    def __init__(self, painter_config: dict, records: list, save_figure=False):
        self.config = painter_config
        self.records = records
        self.save_figure = save_figure

    def generate_obj_option(self, raw_obj_option: dict):
        for k in raw_obj_option:
            if type(raw_obj_option[k]) is list:
                assert len(raw_obj_option[k]) >= len(self.records)
                raw_obj_option[k] = raw_obj_option[k][:len(self.records)]
            else:
                raw_obj_option[k] = [raw_obj_option[k] for _ in self.records]
        return [{k:v[i] for k,v in raw_obj_option.items()} for i in range(len(self.records))]

    def create_figure(self, object_class, fig_config):
        if 'split' in  fig_config.keys():
            cols = fig_config['split']['cols'] if 'cols' in fig_config['split'] else 4
            rows = int(math.ceil(len(self.records)/cols))
            cols = min(len(self.records), cols)
            if 'figsize' in fig_config['split']:
                new_fig_size = (fig_config['split']['figsize'][0], fig_config['split']['figsize'][1])
            else:
                fig_size = mpl.rcParams['figure.figsize']
                new_fig_size = (fig_size[0] * cols, fig_size[1] * rows)
            fig, axs = plt.subplots(rows, cols, figsize=new_fig_size)
            if type(axs) is np.ndarray:
                axs = axs.reshape(-1)
            else:
                axs = [axs]
        else:
            fig, ax = plt.subplots()
            axs = [ax for _ in self.records]
        args = fig_config['args']
        obj_options = self.generate_obj_option(fig_config['obj_option']) if 'obj_option' in fig_config.keys() else [{} for _ in self.records]
        objects = [object_class(rec, args, obj_option) for rec, obj_option in zip(self.records, obj_options)]
        for ob,axi in zip(objects, axs):
            ob.draw(axi)
        if 'fig_option' in fig_config.keys():
            if 'no_legend' in fig_config['fig_option'].keys():
                for obj in objects: obj.with_legend = False
            for option_name in fig_config['fig_option']:
                if option_name=='no_legend': continue
                if 'split' in fig_config.keys():
                    if type(fig_config['fig_option'][option_name]) is str:
                        for ax in axs:
                            eval('ax.set_'+option_name+"('{}')".format(fig_config['fig_option'][option_name]))
                    else:
                        for ax in axs:
                            eval('ax.set_'+option_name+"({})".format(fig_config['fig_option'][option_name]))
                else:
                    if type(fig_config['fig_option'][option_name]) is str:
                        eval('plt.'+option_name+"('{}')".format(fig_config['fig_option'][option_name]))
                    else:
                        eval('plt.' + option_name + "({})".format(fig_config['fig_option'][option_name]))
        plt.show()

    def run(self, group=False):
        for object_class_string in self.config.keys():
            object_class = eval(object_class_string)
            con1 = object_class_string.startswith('Group') and group
            con2 = not object_class_string.startswith('Group') and not group
            if con1 or con2:
                for fig_config in self.config[object_class_string]:
                    self.create_figure(object_class, fig_config)

############################# Table ##############################
def min_value(record,  col_option):
    return np.min(record.data[col_option['x']])

def max_value(record,  col_option):
    return np.max(record.data[col_option['x']])

def variance(record, col_option):
    return np.var(record.data[col_option['x']])

def std_value(record, col_option):
    return np.std(record.data[col_option['x']])

def mean_value(record, col_option):
    return np.mean(record.data[col_option['x']])

def final_value(record, col_option):
    return record.data[col_option['x']][-1]

def optimal_x_by_y(record, col_option):
    if 'flag' not in col_option.keys(): col_option['flag'] = 'min'
    if col_option['flag']=='min': f = np.argmin
    else: f=np.argmax
    tmp = f(record.data[col_option['y']])
    return record.data[col_option['x']][tmp]

def group_optimal_value(record, col_option):
    if 'flag' not in col_option.keys(): col_option['flag'] = 'min'
    if col_option['flag']=='min': f = np.min
    else: f=np.max
    minvs = np.array([f(rdata[col_option['x']]) for rdata in record.datas])
    mean_v = np.mean(minvs)
    std_v = np.std(minvs)
    return "{:.4f} ± {:.4f}".format(mean_v, std_v)

def group_optimal_x_by_y(record, col_option):
    if 'flag' not in col_option.keys(): col_option['flag'] = 'min'
    if col_option['flag']=='min': f = np.argmin
    else: f=np.argmax
    vs = []
    for rdata in record.datas:
        tmp = f(rdata[col_option['y']])
        vs.append(rdata[col_option['x']][tmp])
    mean_v = np.mean(vs)
    std_v = np.std(vs)
    return "{:.4f} ± {:.4f}".format(mean_v, std_v)

class Table:
    def __init__(self, table_config, records, save_text=False):
        self.config = table_config
        self.records = records
        self.save_text = save_text
        self.tb = pt.PrettyTable()
        self.tb.add_column('Record', [r.data['label'] for r in self.records])
        self.sort_key = None

    def set_title(self, title):
        self.tb.title = title

    def create_column(self, funcname, col_option):
        column = []
        for rec in self.records:
            column.append(eval(funcname+'(rec, col_option)'))
        if 'name' in col_option.keys():
            fieldname = col_option['name']
        else:
            fieldname = '-'.join([str(v) for k,v in col_option.items() if k!='sort'])
            fieldname = funcname + '-' + fieldname
        self.tb.add_column(fieldname=fieldname, column=column)
        if 'sort' in col_option.keys(): self.tb.sortby = fieldname

    def run(self, group=False):
        for funcname in self.config:
            con1 = funcname.startswith('group') and group
            con2 = not funcname.startswith('group') and not group
            if con1 or con2:
                col_options = self.config[funcname]
                if type(col_options) is not list: col_options = [col_options]
                for col in col_options:
                    if type(col) is not dict: col_option = {'x': col_option}
                    else: col_option = col
                    self.create_column(funcname, col_option)
        self.tb.float_format = "3.4"
        print(self.tb)

#################################################################
def read_option():
    parser = argparse.ArgumentParser()
    parser.add_argument('--config', help='the configuration of result analysis;', type=str, default='res_config.yml')
    parser.add_argument('--save_figure', help='set True to save the plotted figures', action="store_true", default=False)
    parser.add_argument('--save_text', help='set True to save the printed tables', type=float, default=0)
    parser.add_argument('--seed', help='seed for random initialization;', type=int, default=0)
    try: option = vars(parser.parse_args())
    except IOError as msg: parser.error(str(msg))
    random.seed(option['seed'])
    np.random.seed(option['seed'])
    return option

def show(config, save_figure=False, save_text=False, seed=0):
    option = load_configuration(config)
    # with open(config) as f:
    #     option = yaml.load(f, Loader=yaml.FullLoader)
    record_selector = Selector(option['Selector'])
    if 'Painter' in option.keys():
        for task in record_selector.records:
            p = Painter(option['Painter'], record_selector.records[task])
            p.run()
        for task in record_selector.grouped_records:
            p = Painter(option['Painter'], record_selector.grouped_records[task])
            p.run(group=True)
    if 'Table' in option.keys():
        for task in record_selector.records:
            tb = Table(option['Table'], record_selector.records[task])
            tb.set_title(task)
            tb.run()
        for task in record_selector.grouped_records:
            tb = Table(option['Table'], record_selector.grouped_records[task])
            tb.set_title(task)
            tb.run(group=True)

if __name__ == '__main__':
    option = read_option()
    with open(option['config']) as f:
        cfg = yaml.load(f, Loader=yaml.FullLoader)
    record_selector = Selector(cfg['Selector'])
    if 'Painter' in cfg.keys():
        for task in record_selector.records:
            p = Painter(cfg['Painter'], record_selector.records[task])
            p.run()
        for task in record_selector.grouped_records:
            p = Painter(cfg['Painter'], record_selector.grouped_records[task])
            p.run(group=True)
    if 'Table' in cfg.keys():
        for task in record_selector.records:
            tb = Table(cfg['Table'], record_selector.records[task])
            tb.set_title(task)
            tb.run()
        for task in record_selector.grouped_records:
            tb = Table(cfg['Table'], record_selector.grouped_records[task])
            tb.set_title(task)
            tb.run(group=True)



"""
This is a non-official implementation of 'Fast Federated Learning in the
Presence of Arbitrary Device Unavailability' (http://arxiv.org/abs/2106.04159)
"""
from flgo.algorithm.fedbase import BasicServer, BasicClient
from flgo.algorithm.fedavg import Client
from flgo.utils import fmodule

class Server(BasicServer):
    def initialize(self, *args, **kwargs):
        self.init_algo_para({'c':1.0})
        self.update_table = [None for _ in range(self.num_clients)]
        self.initflag = False
        # choose all the clients that are active
        self.clients_per_round = self.num_clients

    def check_if_init(self):
        """Check whether the update_table is initialized"""
        s = len([u for u in self.update_table if u])
        # c==0 infers that updating starts immediately
        if s < self.c*self.num_clients: return False
        self.gv.logger.info("G_i Initialized For {}/{} The Clients.".format(s,self.num_clients))
        self.initflag = True
        return True

    def iterate(self):
        # sample all the active clients
        self.selected_clients = self.sample()
        # training
        models = self.communicate(self.selected_clients)['model']
        # update G
        for k in range(len(self.received_clients)):
            self.update_table[self.received_clients[k]] = 1.0 / self.lr * (self.model - models[k])
        # check if the update_table being initialized
        if not self.initflag:
            if not self.check_if_init():
                return
        # aggregate: w = w - eta_t * 1/N * sum(G_i)
        self.model = self.aggregate()
        return

    def aggregate(self):
        return self.model - self.lr * fmodule._model_average([update_i for update_i in self.update_table if update_i])


from torch_geometric.datasets import Planetoid
import flgo.benchmark
import os.path
from flgo.benchmark.toolkits.graph.horizontal.link_prediction import *

TaskCalculator = LinkPredicitonTaskCalculator
class TaskGenerator(LinkPredicitonTaskGenenerator):
    def __init__(self, rawdata_path=os.path.join(flgo.benchmark.path,'RAW_DATA', 'PUBMED'),*args, **kwargs):
        super(TaskGenerator, self).__init__(benchmark='pubmed_link_prediction',rawdata_path=rawdata_path,
                                            builtin_class=Planetoid, dataset_name='Pubmed', num_clients=10)

class TaskPipe(LinkPredicitonTaskPipe):
    def __init__(self, task_name):
        super(TaskPipe, self).__init__(task_name, Planetoid, transform=None)


import pynvml
import time
import queue
import sys
import os
import subprocess
import prettytable as pt
import argparse
import yaml
import itertools

pynvml.nvmlInit()
NUM_DEVICES = pynvml.nvmlDeviceGetCount()
# The maximum of the allocated memory(MB) of a gpu to a command
LEAST_FREE_MEMORY = 100
# run commands if device is being available for more than AVAILABLE_INTERVAL (seconds)
AVAILABLE_PERIOD = 30
# check interval
CHECK_INTERVAL = 1
# root path for cmd
COMMAND_WORKING_DIR = os.path.abspath('..')
# root path for bin
BIN_DIR = os.path.dirname(sys.executable)

def read_option():
    parser = argparse.ArgumentParser()
    parser.add_argument('config')
    # basic settings
    parser.add_argument('--m', help='the least memory needed for each command', type=int, default=1500)
    parser.add_argument('--p',  help='the period of confirming the GPU is being adequately free (measured by seconds), default as 30s', type=int, default=5)
    try:
        option = vars(parser.parse_args())
    except IOError as msg:
        parser.error(str(msg))
    global LEAST_FREE_MEMORY, AVAILABLE_PERIOD, CHECK_INTERVAL, BIN_DIR
    # The maximum of the allocated memory(MB) of a gpu to a command
    LEAST_FREE_MEMORY = option['m']
    AVAILABLE_PERIOD = option['p']
    if BIN_DIR[-1] != '/': BIN_DIR += '/'
    with open(option['config']) as f:
        config = yaml.load(f, Loader=yaml.FullLoader)
    return config

def config2cmd(cfg):
    cmds = []
    list_kv = [(k,v) for k,v in cfg.items() if k!='gpu' and type(v) is list]
    list_k = [e[0] for e in list_kv]
    list_v = (e[1] for e in list_kv)
    combination = list(itertools.product(*list_v))
    cmd_components = ['python', 'main.py']
    common_part= ['--'+k+' '+str(v) for k,v in cfg.items() if k!='gpu' and k not in list_k]
    cmd_components.extend(common_part)
    cmd_head = ' '.join(cmd_components)
    for c in combination:
        kv_pair = ['--'+k+' '+str(v) for k,v in zip([key for key in list_k], c)]
        cmd = cmd_head + ' ' + ' '.join(kv_pair)
        cmds.append(cmd)
    # 要改####################################
    if 'gpu' in cfg.keys():
        devices = cfg['gpu']
        if type(devices) is not list: devices = [devices]
        crt_dev_idx = 0
        for ci in range(len(cmds)):
            cmds[ci] = cmds[ci]+' --gpu '+ str(devices[crt_dev_idx])
            crt_dev_idx = (crt_dev_idx+1)%len(devices)
    return cmds

def clear_screen():
    # for windows
    if os.name == 'nt':
        _ = os.system('cls')

    # for mac and linux(here, os.name is 'posix')
    else:
        _ = os.system('clear')


def output_table(dev_list):
    create_time = time.asctime(time.localtime(time.time()))
    tb = pt.PrettyTable()
    tb.title = "Create Time: " + create_time
    tb.field_names = ['Device', 'Finished\\Total', 'Running', "Memory", "Power"]
    clear_screen()
    tb.clear_rows()
    for dm in dev_list:
        if dm.num_commands > 0:
            cmd_state = dm.get_command_state()
            dev_state = dm.get_device_state()
            tb.add_row([dm.name,
                        "{}\\{}".format(cmd_state['Finished'], cmd_state['Total']),
                        cmd_state['Running'],
                        "{}\\{} MB".format(dev_state['allocated'], dev_state['total']),
                        "{}\\{} W".format(dev_state['power_usage'], dev_state['power_limit'])
                        ])
    print(tb)

class DeviceManager:
    def __init__(self, id=-1):
        self.id = id
        self.condition_satisfied = False
        self.cmd_queue = queue.Queue(maxsize=1000)
        self.name = "cuda:{}".format(id) if id > -1 else "cpu"
        self.num_processing = 0
        self.create_time = time.time()
        self.ready_start = time.time()
        self.ready_end = self.ready_start
        self.ready_interval = self.ready_end - self.ready_start
        self.num_finished = 0
        self.num_commands = 0
        self.error_num = 0
        self.pool = set()
        self.error_commands = []
        if id == -1:
            self.using_gpu = False
            self.handle = None
            self.total_memory = 9999999999
            self.free_memory = 9999999999
            self.allocated_memory = 0
            self.power_limit = 99999
            self.power_usage = 0
        else:
            self.using_gpu = True
            self.handle = pynvml.nvmlDeviceGetHandleByIndex(id)
            self.flush_device_info()
        self.max_num_processing = self.total_memory / LEAST_FREE_MEMORY

    def flush(self):
        if self.check_dev_ready() and not self.queue_empty():
            self.run_cmd()
        # check error and finished
        _ = [self.put_cmd(p.args) for p in self.pool if p.poll() is not None and p.poll() != 0]
        self.num_finished += len([p for p in self.pool if p.poll() == 0])
        self.pool.difference_update([p for p in self.pool if p.poll() is not None])
        self.num_processing = len(self.pool)

    def flush_device_info(self):
        if self.using_gpu:
            self.meminfo = pynvml.nvmlDeviceGetMemoryInfo(self.handle)
            self.total_memory = self.byte2mb(self.meminfo.total)
            self.allocated_memory = self.byte2mb(self.meminfo.used)
            self.free_memory = self.byte2mb(self.meminfo.free)
            self.power_limit = pynvml.nvmlDeviceGetEnforcedPowerLimit(self.handle) / 1000
            self.power_usage = pynvml.nvmlDeviceGetPowerUsage(self.handle) / 1000

    def byte2mb(self, size):
        return int(size / 1024 / 1024)

    def get_device_state(self):
        self.flush_device_info()
        return {
            'name': self.name,
            'total': self.total_memory,
            'allocated': self.allocated_memory,
            'free': self.free_memory,
            'power_limit': self.power_limit,
            'power_usage': self.power_usage
        }

    def get_command_state(self):
        return {"Running": self.num_processing, "Total": self.num_commands, "Finished": self.num_finished}

    def condition(self):
        crt_state = self.get_device_state()
        return self.num_processing < self.max_num_processing and crt_state['free'] > LEAST_FREE_MEMORY

    def check_dev_ready(self):
        if self.condition():
            self.ready_end = time.time()
            self.ready_interval = self.ready_end - self.ready_start
        else:
            self.ready_start = time.time()
            self.ready_interval = self.ready_end - self.ready_start
        if self.ready_interval > AVAILABLE_PERIOD:
            return True
        else:
            return False

    def put_cmd(self, cmd):
        if cmd == "": return
        self.cmd_queue.put(cmd)

    def get_cmd(self):
        if self.queue_empty():
            return ""
        return self.cmd_queue.get()

    def queue_empty(self):
        return self.cmd_queue.empty()

    def run_cmd(self):
        cmd = self.get_cmd()
        # 要改####################################
        if cmd != "":
            print("Running cmd: " + cmd)
            self.ready_start = time.time()
            p = subprocess.Popen(cmd, shell=True, cwd=COMMAND_WORKING_DIR, stdout=subprocess.PIPE,
                                 stdin=subprocess.PIPE)
            self.pool.add(p)

# init dev
dev_list = [DeviceManager(i) for i in range(NUM_DEVICES)]
dev_list.append(DeviceManager(-1))
config = read_option()
cmds = config2cmd(config)
num_cmds = len(cmds)
# initialize queues of devices with cmds
num_effective_cmds = 0

# 要改####################################
for cmd in cmds:
    if cmd[-1] == '\n':
        cmd = cmd[:-1]
    cmd = cmd.strip()
    if cmd == "":
        continue
    num_effective_cmds += 1
    cmd = BIN_DIR + cmd
    # choose device
    idx = cmd.find('gpu')
    if idx == -1:
        dev_list[-1].put_cmd(cmd)
    else:
        id = int(cmd[idx + 4])
        dev_list[id].put_cmd(cmd)

for dev in dev_list:
    dev.num_commands = dev.cmd_queue.qsize()

while True:
    num_finished = sum([dm.num_finished for dm in dev_list])
    output_table(dev_list)
    for dm in dev_list:
        dm.flush()
    # 要改####################################
    num_running = sum([d.num_processing for d in dev_list])
    num_unfinished = sum([d.cmd_queue.qsize() for d in dev_list])
    if num_finished >= num_effective_cmds or (num_running == 0 and num_unfinished == 0):
        output_table(dev_list)
        break
    time.sleep(CHECK_INTERVAL)



import community.community_louvain
import torch
from community import community_louvain
from torch import Tensor
from torch_geometric.data import Data, DataLoader
from torch_geometric.datasets import Planetoid
import torch_geometric.transforms as T
from torch_geometric.utils import train_test_split_edges, mask_to_index, index_to_mask, from_networkx

import torch_geometric.utils
import collections
import numpy as np
import os
import json

from flgo.benchmark.toolkits.base import *
import networkx as nx

from flgo.benchmark.toolkits import BasicTaskPipe, BasicTaskCalculator


class GraphClassificationTaskGen(BasicTaskGenerator):
    def __init__(self, benchmark, rawdata_path, builtin_class, dataset_name, transforms=None):
        super(GraphClassificationTaskGen, self).__init__(benchmark, rawdata_path)
        self.builtin_class = builtin_class
        self.dataset_name = dataset_name
        self.transforms = transforms
        self.additional_option = {}

    def load_data(self):
        self.all_data, self.perm = self.builtin_class(root=self.rawdata_path, name=self.dataset_name, transform=self.transforms).shuffle(return_perm=True)
        self.num_samples = len(self.all_data)
        k = int(0.9 * self.num_samples)
        self.train_data = self.all_data[:k]
        self.test_data = list(range(self.num_samples))[k:]

    def partition(self):
        self.local_datas = self.partitioner(self.train_data)
        self.num_clients = len(self.local_datas)


class GraphClassificationTaskPipe(BasicTaskPipe):
    def __init__(self, task_name, buildin_class, transform=None):
        super(GraphClassificationTaskPipe, self).__init__(task_name)
        self.builtin_class = buildin_class
        self.transform = transform

    def save_task(self, generator):
        client_names = self.gen_client_names(len(generator.local_nodes))
        feddata = {'client_names': client_names,
                   'server_data': list(range(len(generator.test_data))),
                   'rawdata_path': generator.rawdata_path,
                   'para': generator.para,
                   'dataset_name': generator.dataset_name,
                   'perm': generator.perm}
        for cid in range(len(client_names)):
            feddata[client_names[cid]] = {'data': generator.local_datas[cid], }
        with open(os.path.join(self.task_path, 'data.json'), 'w') as outf:
            json.dump(feddata, outf)
        return

    def load_data(self, running_time_option) -> dict:
        # load the datasets
        dataset = self.builtin_class(root=self.feddata['rawdata_path'], name=self.feddata['dataset_name'],
                                     transform=self.transform)
        dataset = dataset[self.feddata['perm']]
        test_data = dataset[self.feddata['server_data']]
        server_data_test, server_data_valid = self.split_dataset(test_data, running_time_option['test_holdout'])
        task_data = {'server': {'test': server_data_test, 'valid': server_data_valid}}
        # rearrange data for clients
        for cid, cname in enumerate(self.feddata['client_names']):
            cdata = self.feddata[cname]['data']
            cdata_train, cdata_valid = self.split_dataset(cdata, running_time_option['train_holdout'])
            task_data[cname] = {'train': cdata_train, 'valid': cdata_valid}
        return task_data


class GraphClassificationTaskCalculator(BasicTaskCalculator):
    def __init__(self, device, optimizer_name='sgd'):
        super(GraphClassificationTaskCalculator, self).__init__(device, optimizer_name)
        self.criterion = torch.nn.CrossEntropyLoss()
        self.DataLoader = DataLoader

    def compute_loss(self, model, data):
        """
        :param model: the model to train
        :param data: the training dataset
        :return: dict of train-one-step's result, which should at least contains the key 'loss'
        """
        tdata = self.to_device(data)
        outputs = model(tdata[0])
        loss = self.criterion(outputs, tdata[-1])
        return {'loss': loss}

    @torch.no_grad()
    def test(self, model, dataset, batch_size=64, num_workers=0):
        """
        Metric = [mean_accuracy, mean_loss]
        :param model:
        :param dataset:
        :param batch_size:
        :return: [mean_accuracy, mean_loss]
        """
        model.eval()
        if batch_size == -1: batch_size = len(dataset)
        data_loader = self.get_dataloader(dataset, batch_size=batch_size, num_workers=num_workers)
        total_loss = 0.0
        num_correct = 0
        for batch_id, batch_data in enumerate(data_loader):
            batch_data = self.to_device(batch_data)
            outputs = model(batch_data[0])
            batch_mean_loss = self.criterion(outputs, batch_data[-1]).item()
            y_pred = outputs.data.max(1, keepdim=True)[1]
            correct = y_pred.eq(batch_data[-1].data.view_as(y_pred)).long().cpu().sum()
            num_correct += correct.item()
            total_loss += batch_mean_loss * len(batch_data[-1])
        return {'accuracy': 1.0 * num_correct / len(dataset), 'loss': total_loss / len(dataset)}

    def to_device(self, data):
        return data[0].to(self.device), data[1].to(self.device)

    def get_dataloader(self, dataset, batch_size=64, shuffle=True, num_workers=0):
        if self.DataLoader == None:
            raise NotImplementedError("DataLoader Not Found.")
        return self.DataLoader(dataset, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers)



import matplotlib.pyplot as plt
import random
import matplotlib.colors
import collections
import numpy as np
import os

def visualize_by_class(generator, partitioner):
    all_labels = [d[-1] for d in generator.train_data]
    num_classes = len(set(all_labels))
    ax = plt.subplots()
    colors = [key for key in matplotlib.colors.CSS4_COLORS.keys()]
    random.shuffle(colors)
    client_height = 1
    if hasattr(partitioner, 'dirichlet_dist'):
        client_dist = generator.partitioner.dirichlet_dist.tolist()
        data_columns = [sum(cprop) for cprop in client_dist]
        row_map = {k: i for k, i in zip(np.argsort(data_columns), [_ for _ in range(generator.partitioner.num_parties)])}
        for cid, cprop in enumerate(client_dist):
            offset = 0
            y_bottom = row_map[cid] - client_height / 2.0
            y_top = row_map[cid] + client_height / 2.0
            for lbi in range(len(cprop)):
                plt.fill_between([offset, offset + cprop[lbi]], y_bottom, y_top, facecolor=colors[lbi])
                # plt.barh(cid, cprop[lbi], client_height, left=offset, color=)
                offset += cprop[lbi]
    else:
        data_columns = [len(cidx) for cidx in generator.local_datas]
        row_map = {k: i for k, i in zip(np.argsort(data_columns), [_ for _ in range(generator.partitioner.num_parties)])}
        for cid, cidxs in enumerate(generator.local_datas):
            labels = [int(generator.train_data[did][-1]) for did in cidxs]
            lb_counter = collections.Counter(labels)
            offset = 0
            y_bottom = row_map[cid] - client_height / 2.0
            y_top = row_map[cid] + client_height / 2.0
            for lbi in range(num_classes):
                plt.fill_between([offset, offset + lb_counter[lbi]], y_bottom, y_top, facecolor=colors[lbi])
                offset += lb_counter[lbi]
    plt.xlim(0, max(data_columns))
    plt.ylim(-0.5, generator.partitioner.num_parties - 0.5)
    plt.ylabel('Client ID')
    plt.xlabel('Number of Samples')
    plt.show()



from flgo.system_simulator.base import BasicSimulator
import numpy as np
class Simulator(BasicSimulator):
    def update_client_connectivity(self, client_ids):
        probs = [0.1 for _ in client_ids]
        self.set_variable(client_ids, 'prob_drop', probs)

    def update_client_availability(self):
        self.roundwise_fixed_availability = True
        pa = [0.9 for _ in self.clients]
        pua = [0.1 for _ in self.clients]
        self.set_variable(self.all_clients, 'prob_available', pa)
        self.set_variable(self.all_clients, 'prob_unavailable', pua)

    def update_client_responsiveness(self, client_ids, *args, **kwargs):
        latency = [np.random.randint(5,100) for _ in client_ids]
        self.set_variable(client_ids, 'latency', latency)

    def update_client_completeness(self, client_ids, *args, **kwargs):
        working_amount = [max(int(self.clients[cid].num_steps*np.random.rand()), 1) for cid in client_ids]
        self.set_variable(client_ids, 'working_amount', working_amount)


import torch.nn
from torch.nn import Linear
import torch.nn.functional as F
from torch_geometric.nn import GCNConv
from torch_geometric.nn import global_mean_pool
from flgo.utils.fmodule import FModule

class Model(FModule):
    def __init__(self, hidden_channels=64):
        super(Model, self).__init__()
        self.conv1 = GCNConv(7, hidden_channels)
        self.conv2 = GCNConv(hidden_channels, hidden_channels)
        self.conv3 = GCNConv(hidden_channels, hidden_channels)
        self.lin = Linear(hidden_channels, 2)

    def forward(self, data):
        x, edge_index, batch = data.x, data.edge_index, data.batch
        # 1. Obtain node embeddings
        x = self.conv1(x, edge_index)
        x = x.relu()
        x = self.conv2(x, edge_index)
        x = x.relu()
        x = self.conv3(x, edge_index)

        # 2. Readout layer
        x = global_mean_pool(x, batch)  # [batch_size, hidden_channels]

        # 3. Apply a final classifier
        x = F.dropout(x, p=0.5, training=self.training)
        x = self.lin(x)
        return x

def init_local_module(object):
    pass

def init_global_module(object):
    if 'Server' in object.__class__.__name__:
        object.model = Model().to(object.device)



from torch_geometric.datasets import TUDataset
import flgo.benchmark
import os.path
from flgo.benchmark.toolkits.graph.horizontal.graph_classification import *

TaskCalculator = GraphClassificationTaskCalculator

class TaskGenerator(GraphClassificationTaskGen):
    def __init__(self, rawdata_path=os.path.join(flgo.benchmark.path,'RAW_DATA', 'ENZYMES'),*args, **kwargs):
        super(TaskGenerator, self).__init__(benchmark='enzymes_graph_classification',rawdata_path=rawdata_path,
                                            builtin_class=TUDataset, dataset_name='Enzymes')

class TaskPipe(GraphClassificationTaskPipe):
    def __init__(self, task_name):
        super(TaskPipe, self).__init__(task_name, TUDataset, transform=None)


r"""CIFAR10 is a popular image classification benchmark in federated learning that
is firstly used in `Communication-Efficient Learning of Deep Networks from Decentralized
Data` (https://arxiv.org/abs/1602.05629). To create federated tasks based on this benchmark,
you can try the following example:

Example 1: create a IID partitioned CIFAR10 for 100 clients
>>> config = {'benchmark':{'name':'flgo.benchmark.cifar10_classification'},'partitioner':{'name': 'IIDPartitioner','para':{'num_clients':100}}}
>>> flgo.gen_task(config, task_path='./test_cifar10')
After running the above codes, there will be a task dictionary './test_cifar10' that represents a federated task.
This task is corresponding to a static scenario where the samples in CIFAR10 is I.I.D. allocated to 100 clients,
and is also reusable for different algorithms so that a fair comparison can be easily achieved by optimizing
the same model on the same task.

Example 2: create a non-IID partitioned CIFAR10 for 100 clients
>>> niid_config = {'benchmark':{'name':'flgo.benchmark.cifar10_classification'},'partitioner':{'name': 'DirichletPartitioner','para':{'num_clients':100, 'alpha':0.5}}}
>>> flgo.gen_task(niid_config, task_path='./test_cifar_dir0.5')
To create a non-IID CIFAR10, the only difference is to specify different partitioners in configurations, where
the term `IIDPartitioner` is replaced by `DirichletPartitioner`. In addition, parameters of the partitioner can
be input by `para`.
"""
import flgo.benchmark.toolkits.visualization
from flgo.benchmark.cifar10_classification.model import cnn

default_model = cnn
visualize = flgo.benchmark.toolkits.visualization.visualize_by_class


import torchvision
from flgo.benchmark.toolkits.cv.horizontal.image_classification import BuiltinClassGenerator, BuiltinClassPipe, GeneralCalculator
import flgo.benchmark
import os.path
TaskCalculator = GeneralCalculator

class TaskGenerator(BuiltinClassGenerator):
    def __init__(self, rawdata_path=os.path.join(flgo.benchmark.path,'RAW_DATA', 'EMNIST'), split='byclass'):
        super(TaskGenerator, self).__init__('emnist_classification', rawdata_path, torchvision.datasets.EMNIST, torchvision.transforms.Compose([torchvision.transforms.ToTensor(),]))
        self.split = split

    def load_data(self):
        self.train_data = torchvision.datasets.EMNIST(root=self.rawdata_path, split = self.split, download=True, train=True, transform=self.transform)
        self.test_data = torchvision.datasets.EMNIST(root=self.rawdata_path, split = self.split, download=True, train=False, transform=self.transform)

class TaskPipe(BuiltinClassPipe):
    def __init__(self, task_name):
        super(TaskPipe, self).__init__(task_name, torchvision.datasets.EMNIST, torchvision.transforms.Compose([torchvision.transforms.ToTensor(),]))


from flgo.system_simulator.base import BasicSimulator
import random
import numpy as np
import collections

################################### Initial Availability Mode ##########################################
def ideal_client_availability(simulator, *args, **kwargs):
    probs1 = [1. for _ in simulator.clients]
    probs2 = [0. for _ in simulator.clients]
    simulator.set_variable(simulator.all_clients, 'prob_available', probs1)
    simulator.set_variable(simulator.all_clients, 'prob_unavailable', probs2)
    return

def y_max_first_client_availability(simulator, beta=0.1):
    """
    This setting follows the activity mode in 'Fast Federated Learning in the
    Presence of Arbitrary Device Unavailability' , where each client ci will be ready
    for joining in a round with a static probability:
        pi = alpha * min({label kept by ci}) / max({all labels}) + ( 1 - alpha )
    and the participation of client is independent across rounds. The string mode
    should be like 'YMaxFirst-x' where x should be replaced by a float number.
    """
    # alpha = float(mode[mode.find('-') + 1:]) if mode.find('-') != -1 else 0.1
    def label_counter(dataset):
        return collections.Counter([int(dataset[di][-1]) for di in range(len(dataset))])
    label_num = len(label_counter(simulator.server.test_data))
    probs = []
    for c in simulator.clients:
        c_counter = label_counter(c.train_data + c.valid_data)
        c_label = [lb for lb in c_counter.keys()]
        probs.append((beta * min(c_label) / max(1, label_num - 1)) + (1 - beta))
    simulator.set_variable(simulator.all_clients, 'prob_available', probs)
    simulator.set_variable(simulator.all_clients, 'prob_unavailable', [1 - p for p in probs])
    simulator.roundwise_fixed_availability = True
    return

def more_data_first_client_availability(simulator, beta=0.0001):
    """
    Clients with more data will have a larger active rate at each round.
    e.g. ci=tanh(-|Di| ln(beta+epsilon)), pi=ci/cmax, beta ∈ [0,1)
    """
    p = np.array([len(c.train_data) for c in simulator.server.clients])
    p = p ** beta
    maxp = np.max(p)
    probs = p/maxp
    simulator.set_variable(simulator.all_clients, 'prob_available', probs)
    simulator.set_variable(simulator.all_clients, 'prob_unavailable', [1 - p for p in probs])
    simulator.roundwise_fixed_availability = True

def less_data_first_client_availability(simulator, beta=0.5):
    """
    Clients with less data will have a larger active rate at each round.
            ci=(1-beta)^(-|Di|), pi=ci/cmax, beta ∈ [0,1)
    """
    # alpha = float(mode[mode.find('-') + 1:]) if mode.find('-') != -1 else 0.1
    prop = np.array([len(c.train_data) for c in simulator.server.clients])
    prop = prop ** (-beta)
    maxp = np.max(prop)
    probs = prop/maxp
    simulator.set_variable(simulator.all_clients, 'prob_available', probs)
    simulator.set_variable(simulator.all_clients, 'prob_unavailable', [1 - p for p in probs])
    simulator.roundwise_fixed_availability = True

def y_fewer_first_client_availability(simulator, beta=0.2):
    """
    Clients with fewer kinds of labels will owe a larger active rate.
        ci = |set(Yi)|/|set(Y)|, pi = beta*ci + (1-beta)
    """
    label_num = len(set([int(simulator.server.test_data[di][-1]) for di in range(len(simulator.server.test_data))]))
    probs = []
    for c in simulator.server.clients:
        train_set = set([int(c.train_data[di][-1]) for di in range(len(c.train_data))])
        valid_set = set([int(c.valid_data[di][-1]) for di in range(len(c.valid_data))])
        label_set = train_set.union(valid_set)
        probs.append(beta * len(label_set) / label_num + (1 - beta))
    simulator.set_variable(simulator.all_clients, 'prob_available', probs)
    simulator.set_variable(simulator.all_clients, 'prob_unavailable', [1 - p for p in probs])
    simulator.roundwise_fixed_availability = True

def homogeneous_client_availability(simulator, beta=0.2):
    """
    All the clients share a homogeneous active rate `1-beta` where beta ∈ [0,1)
    """
    # alpha = float(mode[mode.find('-') + 1:]) if mode.find('-') != -1 else 0.8
    probs = [1.-beta for _ in simulator.clients]
    simulator.set_variable(simulator.all_clients, 'prob_available', probs)
    simulator.set_variable(simulator.all_clients, 'prob_unavailable', [1 - p for p in probs])
    simulator.roundwise_fixed_availability = True

def lognormal_client_availability(simulator, beta=0.1):
    """The following two settings are from 'Federated Learning Under Intermittent
    Client Availability and Time-Varying Communication Constraints' (http://arxiv.org/abs/2205.06730).
        ci ~ logmal(0, lognormal(0, -ln(1-beta)), pi=ci/cmax
    """
    epsilon = 0.000001
    Tks = [np.random.lognormal(0, -np.log(1 - beta - epsilon)) for _ in simulator.clients]
    max_Tk = max(Tks)
    probs = np.array(Tks)/max_Tk
    simulator.set_variable(simulator.all_clients, 'prob_available', probs)
    simulator.set_variable(simulator.all_clients, 'prob_unavailable', [1 - p for p in probs])
    simulator.roundwise_fixed_availability = True

def sin_lognormal_client_availability(simulator, beta=0.1):
    """This setting shares the same active rate distribution with LogNormal, however, the active rates are
    also influenced by the time (i.e. communication round). The active rates obey a sin wave according to the
    time with period T.
        ci ~ logmal(0, lognormal(0, -ln(1-beta)), pi=ci/cmax, p(i,t)=(0.4sin((1+R%T)/T*2pi)+0.5) * pi
    """
    # beta = float(mode[mode.find('-') + 1:]) if mode.find('-') != -1 else 0.1
    epsilon = 0.000001
    Tks = [np.random.lognormal(0, -np.log(1 - beta - epsilon)) for _ in simulator.clients]
    max_Tk = max(Tks)
    q = np.array(Tks)/max_Tk
    simulator.set_variable(simulator.all_clients, 'q', q)
    simulator.set_variable(simulator.all_clients, 'prob_available', q)
    def f(self):
        T = 24
        times = np.linspace(start=0, stop=2 * np.pi, num=T)
        fts = 0.4 * np.sin(times) + 0.5
        t = self.server.current_round % T
        q = self.get_variable(self.all_clients, 'q')
        probs = [fts[t]*qi for qi in q]
        self.set_variable(self.all_clients, 'prob_available', probs)
        self.set_variable(self.all_clients, 'prob_unavailable', [1 - p for p in probs])
    simulator.roundwise_fixed_availability = True
    return f

def y_cycle_client_availability(simulator, beta=0.5):
    # beta = float(mode[mode.find('-') + 1:]) if mode.find('-') != -1 else 0.5
    max_label = max(set([int(simulator.server.test_data[di][-1]) for di in range(len(simulator.server.test_data))]))
    for c in simulator.clients:
        train_set = set([int(c.train_data[di][-1]) for di in range(len(c.train_data))])
        valid_set = set([int(c.valid_data[di][-1]) for di in range(len(c.valid_data))])
        label_set = train_set.union(valid_set)
        c._min_label = min(label_set)
        c._max_label = max(label_set)
    def f(self):
        T = 24
        r = 1.0 * (1 + self.server.current_round % T) / T
        probs = []
        for c in self.clients:
            ic = int(r >= (1.0 * c._min_label / max_label) and r <= (1.0 * c._max_label / max_label))
            probs.append(beta * ic + (1 - beta))
        self.set_variable(self.all_clients, 'prob_available', probs)
        self.set_variable(self.all_clients, 'prob_unavailable', [1 - p for p in probs])
    simulator.roundwise_fixed_availability = True
    return f

################################### Initial Connectivity Mode ##########################################
def ideal_client_connectivity(simulator, *args, **kwargs):
    probs = [0. for _ in simulator.clients]
    simulator.set_variable(simulator.all_clients, 'prob_drop', probs)

def homogeneous_client_connectivity(simulator, gamma=0.05):
    probs = [gamma for _ in simulator.clients]
    simulator.set_variable(simulator.all_clients, 'prob_drop', probs)

################################### Initial Completeness Mode ##########################################
def ideal_client_completeness(simulator, *args, **kwargs):
    simulator.set_variable(simulator.all_clients, 'working_amount', [c.num_steps for c in simulator.clients])
    return

def part_dynamic_uniform_client_completeness(simulator, p=0.5):
    """
    This setting follows the setting in the paper 'Federated Optimization in Heterogeneous Networks'
    (http://arxiv.org/abs/1812.06127). The `p` specifies the number of selected clients with
    incomplete updates.
    """
    simulator.prob_incomplete = p
    def f(self, client_ids = []):
        was = []
        for cid in client_ids:
            wa = self.random_module.randint(low=0, high=self.clients[cid].num_steps) if self.random_module.rand() < self.prob_incomplete else self.clients[cid].num_steps
            wa = max(1, wa)
            was.append(wa)
            self.clients[cid].num_steps = wa
        self.set_variable(client_ids, 'working_amount', was)
        return
    return f

def full_static_unifrom_client_completeness(simulator):
    working_amounts = [max(1, int(c.num_steps * np.random.rand())) for c in simulator.clients]
    simulator.set_variable(simulator.all_clients, 'working_amount', working_amounts)
    return

def arbitrary_dynamic_unifrom_client_completeness(simulator, a=1, b=1):
    """
    This setting follows the setting in the paper 'Tackling the Objective Inconsistency Problem in
    Heterogeneous Federated Optimization' (http://arxiv.org/abs/2007.07481). The string `mode` should be like
    'FEDNOVA-Uniform(a,b)' where `a` is the minimal value of the number of local epochs and `b` is the maximal
    value. If this mode is active, the `num_epochs` and `num_steps` of clients will be disable.
    """
    simulator._incomplete_a = min(a, 1)
    simulator._incomplete_b = max(b, simulator._incomplete_a)
    def f(self, client_ids = []):
        for cid in client_ids:
            self.clients[cid].set_local_epochs(self.random_module.randint(low=self._incomplete_a, high=self._incomplete_b))
        working_amounts = [self.clients[cid].num_steps for cid in self.all_clients]
        self.set_variable(self.all_clients, 'working_amount', working_amounts)
        return
    return f

def arbitrary_static_unifrom_client_completeness(simulator, a=1, b=1):
    """
    This setting follows the setting in the paper 'Tackling the Objective Inconsistency Problem in
    Heterogeneous Federated Optimization' (http://arxiv.org/abs/2007.07481). The string `mode` should be like
    'FEDNOVA-Uniform(a,b)' where `a` is the minimal value of the number of local epochs and `b` is the maximal
    value. If this mode is active, the `num_epochs` and `num_steps` of clients will be disable.
    """
    a = min(a, 1)
    b = max(b, a)
    for cid in simulator.clients:
        simulator.clients[cid].set_local_epochs(np.random.randint(low=a, high=b))
    working_amounts = [simulator.clients[cid].num_steps for cid in simulator.all_clients]
    simulator.set_variable(simulator.all_clients, 'working_amount', working_amounts)
    return

################################### Initial Timeliness Mode ############################################
def ideal_client_responsiveness(simulator, *args, **kwargs):
    latency = [0 for _ in simulator.clients]
    for c, lt in zip(simulator.clients, latency): c._latency = lt
    simulator.set_variable(simulator.all_clients, 'latency', latency)

def lognormal_client_responsiveness(simulator, mean_latency=100, var_latency=10):
    mu = np.log(mean_latency) - 0.5 * np.log(1 + var_latency / mean_latency / mean_latency)
    sigma = np.sqrt(np.log(1 + var_latency / mean_latency / mean_latency))
    client_latency = np.random.lognormal(mu, sigma, len(simulator.clients))
    latency = [int(ct) for ct in client_latency]
    for c, lt in zip(simulator.clients, latency): c._latency = lt
    simulator.set_variable(simulator.all_clients, 'latency', latency)

def uniform_client_responsiveness(simulator, min_latency=0, max_latency=1):
    latency = [np.random.randint(low=min_latency, high=max_latency) for _ in simulator.clients]
    for c,lt in zip(simulator.clients, latency): c._latency = lt
    simulator.set_variable(simulator.all_clients, 'latency', latency)

#************************************************************************************************
availability_modes = {
    'IDL': ideal_client_availability,
    'YMF': y_max_first_client_availability,
    'MDF': more_data_first_client_availability,
    'LDF': less_data_first_client_availability,
    'YFF': y_fewer_first_client_availability,
    'HOMO': homogeneous_client_availability,
    'LN': lognormal_client_availability,
    'SLN': sin_lognormal_client_availability,
    'YC': y_cycle_client_availability,
}

connectivity_modes = {
    'IDL': ideal_client_connectivity,
    'HOMO': homogeneous_client_connectivity,
}

completeness_modes = {
    'IDL': ideal_client_completeness,
    'PDU': part_dynamic_uniform_client_completeness,
    'FSU': full_static_unifrom_client_completeness,
    'ADU': arbitrary_dynamic_unifrom_client_completeness,
    'ASU': arbitrary_static_unifrom_client_completeness,
}

responsiveness_modes = {
    'IDL': ideal_client_responsiveness,
    'LN': lognormal_client_responsiveness,
    'UNI': uniform_client_responsiveness,
}

class Simulator(BasicSimulator):
    def __init__(self, objects, option = None):
        super().__init__(objects)
        self.option = option
        # +++++++++++++++++++++ availability +++++++++++++++++++++
        avl_mode, avl_para = self.get_mode(option['availability'])
        if avl_mode not in availability_modes: avl_mode, avl_para = 'IDL', ()
        f_avl = availability_modes[avl_mode](self, *avl_para)
        if f_avl is not None: self.__class__.update_client_availability = f_avl
        # +++++++++++++++++++++ connectivity +++++++++++++++++++++
        con_mode, con_para = self.get_mode(option['connectivity'])
        if con_mode not in connectivity_modes: con_mode, con_para = 'IDL', ()
        f_con = connectivity_modes[con_mode](self, *con_para)
        if f_con is not None: self.__class__.update_client_connectivity = f_con
        # +++++++++++++++++++++ completeness +++++++++++++++++++++
        cmp_mode, cmp_para = self.get_mode(option['completeness'])
        if cmp_mode not in completeness_modes: cmp_mode, cmp_para = 'IDL', ()
        f_cmp = completeness_modes[cmp_mode](self, *cmp_para)
        if f_cmp is not None: self.__class__.update_client_completeness = f_cmp
        # +++++++++++++++++++++ responsiveness ++++++++++++++++++++++++
        rsp_mode, rsp_para = self.get_mode(option['responsiveness'])
        if rsp_mode not in responsiveness_modes: rsp_mode, rsp_para = 'IDL', ()
        f_rsp = responsiveness_modes[rsp_mode](self, *rsp_para)
        if f_rsp is not None: self.__class__.update_client_responsiveness = f_rsp
        if self.server.tolerance_for_latency == 0:
            self.server.tolerance_for_latency = max([c._latency for c in self.clients])
        return

    def get_mode(self, mode_string):
        mode = mode_string.split('-')
        mode, para = mode[0], mode[1:]
        if len(para) > 0: para = [float(pi) for pi in para]
        return mode, tuple(para)


import importlib
import shutil
from abc import ABCMeta, abstractmethod
import random
import matplotlib.pyplot as plt
import numpy as np
import os
import torch
from torch.utils.data import Dataset
import json

class AbstractTaskGenerator(metaclass=ABCMeta):
    @abstractmethod
    def load_data(self, *args, **kwarg):
        """Load the original data into memory that can be partitioned"""
        pass

    @abstractmethod
    def partition(self, *args, **kwarg):
        """Partition the loaded data into subsets of data owned by clients
        and the test data owned by the server
        """
        pass

    @abstractmethod
    def generate(self, *args, **kwarg):
        """Load and partition the data, and then generate the necessary
        information about the federated task (e.g. path, partition way, ...)"""
        pass


class AbstractTaskPipe(metaclass=ABCMeta):
    @abstractmethod
    def save_task(self, *args, **kwargs):
        """Save a federated task created by TaskGenerator as a static file on the disk"""
        pass

    @abstractmethod
    def load_task(self, *args, **kwargs):
        """Load a federated task from disk"""
        pass


class AbstractTaskCalculator(metaclass=ABCMeta):
    @abstractmethod
    def to_device(self, *args, **kwargs):
        pass

    @abstractmethod
    def get_dataloader(self, *args, **kwargs):
        pass

    @abstractmethod
    def test(self, model, data, *args, **kwargs):
        pass

    @abstractmethod
    def compute_loss(self, model, data, *args, **kwargs):
        pass

    @abstractmethod
    def get_optimizer(self, model, *args, **kwargs):
        pass

class BasicTaskGenerator(AbstractTaskGenerator):
    def __init__(self, benchmark, rawdata_path):
        """
        :param benchmark: the name of the ML task to be converted
        :param rawdata_path: the dictionary of the original dataset
        """
        # basic attribution
        self.benchmark = benchmark
        self.rawdata_path = rawdata_path
        # optional attribution
        self.partitioner = None
        self.train_data = None
        self.test_data = None
        self.task_name = None
        self.para = {}

    def generate(self, *args, **kwarg):
        """The whole process to generate federated task. """
        # load data
        self.load_data()
        # partition
        self.partition()
        # generate task name
        self.task_name = self.get_task_name()
        return

    def load_data(self, *args, **kwargs):
        """Download and load dataset into memory."""
        raise NotImplementedError

    def partition(self, *args, **kwargs):
        """Partition the data"""
        return

    def register_partitioner(self, partitioner=None):
        self.partitioner = partitioner

    def init_para(self, para_list=None):
        pnames = list(self.para.keys())
        if para_list is not None:
            for i, pv in enumerate(para_list):
                pname = pnames[i]
                try:
                    self.para[pname] = type(self.para[pname])(pv)
                except:
                    self.para[pname] = pv
        for pname, pv in self.para.items():
            self.__setattr__(pname, pv)
        return

    def get_task_name(self):
        if not hasattr(self.partitioner, 'num_parties') and hasattr(self.partitioner, 'num_clients'):
            self.partitioner.num_parties = self.partitioner.num_clients
        else: self.partitioner.num_parties = 'unknown'
        return '_'.join(['B-' + self.benchmark, 'P-' + str(self.partitioner), 'N-' + str(self.partitioner.num_parties)])

class BasicTaskPipe(AbstractTaskPipe):
    TaskDataset = None

    def __init__(self, task_path):
        self.task_path = task_path
        if os.path.exists(os.path.join(self.task_path, 'data.json')):
            with open(os.path.join(self.task_path, 'data.json'), 'r') as inf:
                self.feddata = json.load(inf)

    def save_task(self, generator):
        # Construct `feddata` and store it into the disk for recover the partitioned datasets again from it
        raise NotImplementedError

    def load_data(self, running_time_option) -> dict:
        # Load the data and process it to the format that can be distributed to different objects
        raise NotImplementedError

    def generate_objects(self, running_time_option, algorithm, scene='horizontal') -> list:
        # Generate the virtual objects (i.e. coordinators and participants) in the FL system
        if scene=='horizontal':
            # init clients
            Client = algorithm.Client
            clients = [Client(running_time_option) for _ in range(len(self.feddata['client_names']))]
            for cid, c in enumerate(clients):
                c.id = cid
                c.name = self.feddata['client_names'][cid]
            # init server
            server = algorithm.Server(running_time_option)
            server.name = 'server'
            # bind clients and server
            server.register_clients(clients)
            for c in clients: c.register_server(server)
            # return objects as list
            objects = [server]
            objects.extend(clients)
        elif scene=='vertical':
            PassiveParty = algorithm.PassiveParty
            ActiveParty = algorithm.ActiveParty
            objects = []
            for pid, pname in enumerate(self.feddata['party_names']):
                is_active = self.feddata[pname]['data']['with_label']
                obj = ActiveParty(running_time_option) if is_active else PassiveParty(running_time_option)
                obj.id = pid
                obj.name = pname
                objects.append(obj)
            for party in objects:
                party.register_parties(objects)
        return objects

    def save_info(self, generator):
        info = {'benchmark': '.'.join(generator.__module__.split('.')[:-1])}
        info['scene'] = generator.scene if hasattr(generator, 'scene') else 'unknown'
        info['num_clients'] = generator.num_clients if hasattr(generator, 'num_clients') else (generator.num_parties if hasattr(self, 'num_parties') else 'unknown')
        with open(os.path.join(self.task_path, 'info'), 'w') as outf:
            json.dump(info, outf)

    def load_task(self, running_time_option, *args, **kwargs):
        task_data = self.load_data(running_time_option)
        objects = self.generate_objects(running_time_option)
        self.distribute(task_data, objects)
        return objects

    def distribute(self, task_data: dict, objects: list):
        for ob in objects:
            ob_data = task_data[ob.name]
            for data_name, data in ob_data.items():
                ob.set_data(data, data_name)

    def split_dataset(self, dataset, p=0.0):
        if p == 0: return dataset, None
        s1 = int(len(dataset) * p)
        s2 = len(dataset) - s1
        return torch.utils.data.random_split(dataset, [s2, s1])

    def task_exists(self):
        """Check whether the task already exists."""
        return os.path.exists(self.task_path)

    def remove_task(self):
        "remove the task"
        if self.task_exists():
            shutil.rmtree(self.task_path)
        return

    def create_task_architecture(self):
        """Create the directories of the task."""
        if not self.task_exists():
            os.mkdir(self.task_path)
            os.mkdir(os.path.join(self.task_path, 'record'))
            os.mkdir(os.path.join(self.task_path, 'log'))
        else:
            raise FileExistsError("federated task {} already exists!".format(self.task_path))

    def save_figure(self):
        plt.savefig(os.path.join(self.task_path, 'res.png'))

    def gen_client_names(self, num_clients):
        return [('Client{:0>' + str(len(str(num_clients))) + 'd}').format(i) for i in range(num_clients)]


class BasicTaskCalculator(AbstractTaskCalculator):
    def __init__(self, device, optimizer_name='sgd'):
        self.device = device
        self.optimizer_name = optimizer_name
        self.criterion = None
        self.DataLoader = None

    def to_device(self, data, *args, **kwargs):
        return NotImplementedError

    def get_dataloader(self, *args, **kwargs):
        return NotImplementedError

    def test(self, model, data, *args, **kwargs):
        return NotImplementedError

    def compute_loss(self, model, data, *args, **kwargs):
        return NotImplementedError

    def get_optimizer(self, model=None, lr=0.1, weight_decay=0, momentum=0):
        OPTIM = getattr(importlib.import_module('torch.optim'), self.optimizer_name)
        filter_fn = filter(lambda p: p.requires_grad, model.parameters())
        if self.optimizer_name.lower() == 'sgd':
            return OPTIM(filter_fn, lr=lr, momentum=momentum, weight_decay=weight_decay)
        elif self.optimizer_name.lower() in ['adam', 'rmsprop', 'adagrad']:
            return OPTIM(filter_fn, lr=lr, weight_decay=weight_decay)
        else:
            raise RuntimeError("Invalid Optimizer.")


# class HorizontalTaskPipe(BasicTaskPipe):
#     def generate_objects(self, running_time_option):
#         # init clients
#         client_path = '%s.%s' % ('algorithm', running_time_option['algorithm'])
#         Client = getattr(importlib.import_module(client_path), 'Client')
#         clients = [Client(running_time_option) for _ in range(len(self.feddata['client_names']))]
#         for cid, c in enumerate(clients):
#             c.id = cid
#             c.name = self.feddata['client_names'][cid]
#         # init server
#         server_path = '%s.%s' % ('algorithm', running_time_option['algorithm'])
#         server_module = importlib.import_module(server_path)
#         server = getattr(server_module, 'Server')(running_time_option)
#         server.name = 'server'
#         server.register_clients(clients)
#         for c in clients: c.register_server(server)
#         objects = [server]
#         objects.extend(clients)
#         return objects


class XYHorizontalTaskPipe(BasicTaskPipe):
    """
    This pipe is for supervised learning where each sample contains a feature $x_i$ and a label $y_i$
     that can be indexed by $i$.
    To use this pipe, it's necessary to set the attribute `test_data` of the generator to be a dict like:
        {'x': [...], 'y':[...]}
    and the attribute `local_datas` to be a list of the above dict that means the local data owned by clients:
        [{'x':[...], 'y':[...]}, ..., ]
    """
    TaskDataset = torch.utils.data.TensorDataset

    def save_task(self, generator):
        client_names = self.gen_client_names(len(generator.local_datas))
        feddata = {'client_names': client_names, 'server': {'data': generator.test_data}}
        for cid in range(len(client_names)): feddata[client_names[cid]] = {'data': generator.local_datas[cid]}
        with open(os.path.join(self.task_path, 'data.json'), 'w') as outf:
            json.dump(feddata, outf)

    def load_data(self, running_time_option) -> dict:
        test_data = self.feddata['server']['data']
        test_data = self.TaskDataset(torch.tensor(test_data['x']), torch.tensor(test_data['y']))
        local_datas = [self.TaskDataset(torch.tensor(self.feddata[cname]['data']['x']),
                                        torch.tensor(self.feddata[cname]['data']['y'])) for cname in
                       self.feddata['client_names']]
        server_data_test, server_data_valid = self.split_dataset(test_data, running_time_option['test_holdout'])
        task_data = {'server': {'test': server_data_test, 'valid': server_data_valid}}
        for key in self.feddata['server'].keys():
            if key == 'data':
                continue
            task_data['server'][key] = self.feddata['server'][key]
        for cid, cname in enumerate(self.feddata['client_names']):
            cdata = local_datas[cid]
            cdata_train, cdata_valid = self.split_dataset(cdata, running_time_option['train_holdout'])
            task_data[cname] = {'train': cdata_train, 'valid': cdata_valid}
            for key in self.feddata[cname]:
                if key == 'data':
                    continue
                task_data[cname][key] = self.feddata[cname][key]
        return task_data

# class IDXHorizontalTaskPipe(HorizontalTaskPipe):
#     TaskDataset = torch.utils.data.Subset
#     def save_task(self, generator):
#         client_names = self.gen_client_names(len(generator.local_datas))
#         feddata = {'client_names': client_names, 'server': {'data': generator.test_data}}
#         pass



"""
This is a non-official implementation of the work 'TiFL: A Tier-based Federated Learning System'
(http://arxiv.org/abs/2001.09249). This implementation refers to the github repository https://github.com/bacox/fltk.
Thanks for their great work.

T: the number of tiers
I: the interval of updating testing accuracy for each tier
C: the degree of the diverse in the `Credits` for tiers. The smaller C is, the uniform the Credits distributes.
"""
from .fedbase import BasicServer
from .fedbase import BasicClient as Client
import numpy as np

class Server(BasicServer):
    def initialize(self):
        self.init_algo_para({'T': 5, 'I': 10, 'C': 0})
        self.tolerance_for_latency = 0
        self.tiers = None

    def profiling_and_tiering(self):
        # collecting the latency for each client
        client_latencies = [c._latency for c in self.clients]
        ordered_cids = np.argsort(client_latencies)
        groups = np.split(ordered_cids, self.T)
        group_latency = [(np.array(client_latencies)[group]).mean() for group in groups]
        tiers = {tid: {
            'clients': groups[tid],
            'ave_latency': group_latency[tid],
            'accuracy':0.0,
            'prob':1.0/self.T,
            'credit': self.num_rounds,
        } for tid in range(self.T)}
        p = (-4 * np.log(self.C + 10e-8)) ** 4
        credits = np.random.dirichlet([p for _ in tiers])
        credits = np.around(credits, 3)
        credits = credits/credits.sum()
        credits.sort()
        rounds = np.ones(self.num_rounds)
        credits = np.split(rounds, (np.cumsum(credits)*self.num_rounds).astype(int)[:-1])
        credits = [ci.sum() for ci in credits]
        tids = [tids for tids in tiers]
        tids = sorted(tids, key=lambda x:  tiers[x]['ave_latency'], reverse=True)
        for tid, c in zip(tids, credits): tiers[tid]['credit'] = int(c)
        return tiers

    def iterate(self):
        if self.tiers is None: self.tiers =  self.profiling_and_tiering()
        if (self.current_round-1)%self.I==0 and self.current_round-1>=self.I:
            self.change_probs()
        self.selected_clients = self.sample()
        models = self.communicate(self.selected_clients)['model']
        self.model = self.aggregate(models)
        # update accuracy of each tier
        valid_accs = np.array(self.global_test('valid')['accuracy'])
        for t in self.tiers:
            self.tiers[t]['accuracy'] = (valid_accs[self.tiers[t]['clients']]).mean()
        return

    def sample(self):
        probs = [self.tiers[t]['prob'] for t in self.tiers.keys()]
        while True:
            self.current_tier = int(np.random.choice(list(self.tiers.keys()), 1, p=probs))
            if self.tiers[self.current_tier]['credit'] > 0:
                self.tiers[self.current_tier]['credit'] = self.tiers[self.current_tier]['credit']-1
                break
        selected_clients = np.random.choice(self.tiers[self.current_tier]['clients'], min(self.clients_per_round, len(self.tiers[self.current_tier]['clients'])))
        return selected_clients.tolist()

    def change_probs(self):
        n = len([t for t in self.tiers if self.tiers[t]['credit']>0])
        D = n * (n + 1) / 2
        tier_accs = [self.tiers[t]['accuracy'] for t in self.tiers]
        sorted_tid = np.argsort(tier_accs)
        tiers = np.array([t for t in self.tiers])[sorted_tid]
        idx_decr = 0
        for idx, t in enumerate(tiers):
            if self.tiers[t]['credit'] > 0:
                self.tiers[t]['prob'] = (n - (idx - idx_decr)) / D
            else:
                self.tiers[t]['prob'] = 0
                idx_decr += 1
        return
